/* Standard-Definitions */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <limits.h>
#include <float.h>
/* Definitions for that program */
#include "useful.h"
#include "complex.h"
#include "matrix.h"
#include "vectors.h"
#include "apply_h.h"

/* The parameters */

#define DELTA 0.9878982465L
#define DELTAP 1.0L
double jval = (double) -1;
double jpval = (double) 1;

/* Two special numbers (should be negative and different, otherwise arbitrary) */

#define MAGIC -12345
#define MAGIX -87654

/* The maximal linear extent */

#define MAXD 32

/* The size in x-direction and y-direction */

static int xsize=9;
static int ysize=MAGIC;

/* The total number of sites */

static int sites;

/* The number of particles (corresponding to total S^z) */

static int n_part=MAGIC;

/* The variours parities */

static int momentum_y=1;	/* y-momentum */
static int parity_diag=1;	/* reflection diagonal */
static int parity_y=1;		/* reflection y */

/* The number of states in the total-S^z vector space */

static long nstates;

/* Dimension of the momentum eigenspace */

static long pstates;

/* Neighbour pairs */

#define NNEIGH 27

struct neighbours {char l; char r; int lm; int rm; };

static struct neighbours neighbours[NNEIGH] =
   { {1,2}, {4,5}, {7,8},
     {2,3}, {5,6}, {8,9},
     {3,1}, {6,4}, {9,7},
     {1,4}, {4,7}, {7,1},
     {2,5}, {5,8}, {8,2},
     {3,6}, {6,9}, {9,3},
     {4,2}, {7,5}, {1,8},
     {5,3}, {8,6}, {2,9},
     {6,1}, {9,4}, {3,7}
   };

/* Mapping tables */

static long *charge_momentum;
	/* Which S^z eigenvector belongs to which momentum eigenvector ? */

static int *charge_prefac;
	/* The (sign) prefactors */

static long *momentum_charge;
	/* Representative S^z eigenvector for the momentum eigenvector */

static int *momentum_prefac;
	/* The number of states associated to the momentum eigenvector */

/* Mapping tables */

#ifndef SLOWINDEX
long ***index_tab;
char *np_tab;

int tab_size;
#endif

/* Table of binomial coefficients */

static long *bin_coeff[1200];

/* ****** Some math routines ****** */

/* define DIRECTBIN in order not to store, but to compute the binomial
   coefficients - useful for very large systems at small densities  */

#ifdef DIRECTBIN
  #define over(a,b) binomial(a,b)
#else
  #define over(a,b) bin_coeff[a][b]
#endif

/* Den groessten gemeinsamen Teiler zweier Zahlen feststellen
   ueber Euklidischen Algorithmus
   Zehnerpotenzen schneller als Primfaktorzerlegung */

static long get_gcd(long a, long b)
 {
   register int i;
   long remainder;

   if(a == 0)           /* Sonderbehandlung fuer Nullen */
     return(1);
   if(b == 0)
     return(1);

   if(a < 0)            /* Betrag bilden */
     a=-a;
   if(b < 0)
     b=-b;

   if(a == 1)           /* Sonderbehandlung fuer Einsen */
     return(1);
   if(b == 1)
     return(1);

   remainder = a % b;   /* Das ist der Euklische Algorithmus */
   while(remainder)
    {
      a=b;
      b=remainder;
      remainder = a % b;
    }           /* Nun steht der GGT in b */

   return(b);
 }

/* Der Binomialkoeffizient */

static long binomial(long n, long m)
 {
   register long i;
   register long gcd;
   register long res1, res2;

   if(m > n-m)                  /* Symmetrie ausnutzen */
     m = n-m;

   if(m < 0)
     return(0);

   res1 = 1;
   res2 = 1;
   for(i=0; i<m; i++)
    {
      res1 *= (n-i);           /* Multiplizieren */
      res2 *= (i+1);
      gcd = get_gcd(res1, res2);      /* Und kuerzen */
      res1 /= gcd;
      res2 /= gcd;
    }
   if(res2 != 1)
     fprintf(stderr, "*** Error: Binomial coefficient (%ld,%ld) not an integer !\n", n, m);
   return(res1);
 }

/* Allocate memory and produce error message if not available */

static void *m_alloc(long size, char *str)
 {
  void *ptr;

  ptr = (void *)malloc(size);

  if(! ptr)
   {
    fprintf(stderr, "### Error: Out of memory in %s()\n", str);
    exit(-13);
   }
  return(ptr);
 }

/* Initialize binomial coefficients */

static void init_binomials()
 {
  register long n, m;

  for(n=0; n<=sites; n++)
   {
    bin_coeff[n] = (long *)m_alloc((((n+1)<n_part)?(n+2):(n_part+1))*sizeof(long), "init_binomials");
    for(m=0; m<=n+1; m++)
     if(m <= n_part)
       bin_coeff[n][m] = binomial(n,m);
   }
 }

/* Free binomial coefficients */

static void free_binomials()
 {
  register long n;

  for(n=0; n<=sites; n++)
    free(bin_coeff[n]);
 }

#ifndef SLOWINDEX

static char IMSTR[] = "init_mtables";

/* Initialize mapping tables */

static void init_mtables()
 {
  register int i, j, x;
  long mask, index;
  int cnt, np, lnp;

  tab_size = 1<<xsize;

  np_tab = (char *)m_alloc(tab_size * sizeof(char), IMSTR);

  for(i=0; i<tab_size; i++)
   {
    cnt = 0;
    mask = 1;
    for(x = 0; x<xsize; x++)
     {
       if(i & mask)
         cnt++;
       mask = mask<<1;
     }
    np = np_tab[i] = cnt;
   }

  index_tab = (long ***)m_alloc(tab_size * sizeof(long **), IMSTR);
  for(i=0; i<tab_size; i++)
   {
    index_tab[i] = (long **)m_alloc(ysize * sizeof(long *), IMSTR);
    for(j=0; j<ysize; j++)
     {
      if(n_part - ((int) np_tab[i]) >= 0)
        index_tab[i][j] = (long *)m_alloc((n_part - np_tab[i] + 1)
                  * sizeof(long), IMSTR);
      else
        index_tab[i][j] = 0;
     }
   }

  for(i=0; i<tab_size; i++)
   for(j=0; j<ysize; j++)
    for(lnp=0; lnp<=n_part - ((int) np_tab[i]); lnp++)
     {
      np = lnp + np_tab[i];
      cnt = j * xsize;
      index = 0;
      mask = 1<<(xsize-1);
      for(x=0; x<xsize; x++)
       {
        if(i & mask)
         {
          index += over(sites-cnt-1, np);
          np--;
          if(np < 0)
           fprintf(stderr, "### Error: Illegal bit pattern %x encountered:\n", i);
         }
        cnt++;
        mask = mask>>1;
       }
      index_tab[i][j][lnp] = index; 
     }
 }

/* Free mapping tables */

static void free_mtables()
 {
  register int i, j;

  for(i=0; i<tab_size; i++)
   {
    for(j=0; j<ysize; j++)
       if(index_tab[i][j])
           free(index_tab[i][j]);
    free(index_tab[i]);
   }
  free(index_tab);
  free(np_tab);
 }
#endif

/* Generate states in S^z-symmetry subspace */

static void gen_states()
 {
  nstates = over(sites, n_part);
 }

/* Free states in S^z-symmetry subspace */

static void free_states()
 {
 }

/* Convert index to bit representation of state */

static unsigned int *index_to_state(long index)
 {
  long oindex;
  unsigned int *state;
  unsigned int mask;
  int np, hh;
  int count;
  register int i, j, k;

  state = (unsigned int *) m_alloc(sizeof(unsigned int) * ysize, "index_to_state");
  for(j=0; j<ysize; j++)
    state[j] = 0;

  oindex = index;

  count = 0;
  np = n_part;
  for(j=0; j<ysize; j++)
   {
    mask = 1<<(xsize-1);
    for(i=0; i<xsize; i++)
     {
      hh = over(sites-count-1, np);
#ifdef DEBUG
      printf("(%d,%d) = %d\n", sites-count-1, np, hh);
#endif
      if(index >= hh)
       {
        index -= hh;
        np--;
        state[j] |= mask;
       }
      count++;
      mask = mask>>1;
     }
   }

  if(index || (np < 0))
   fprintf(stderr, "*** Error: failed to translate index `%ld' to state\n", oindex);

#ifdef DEBUG
  fprintf(stderr, "Result of index_to_state:\n");
  for(k=0; k<ysize; k++)
     fprintf(stderr, "\t%lx\n", state[k]);
#endif

  return(state);
 }

/* Convert bit representation of state to index of matrix */

static long state_to_index(unsigned int *state)
 {
  long index;
#ifdef SLOWINDEX
  unsigned int mask;
  int count;
#else
  int lnp;
#endif
  int np;
  register int i, j, k;

#ifdef SLOWINDEX
  count = index = 0;
  np = n_part;
  for(j=0; j<ysize; j++)
   {
    mask = 1<<(xsize-1);
    for(i=0; i<xsize; i++)
     {
      if(state[j] & mask)
       {
        index += over(sites-count-1, np);
        np--;
        if(np < 0)
         {
          fprintf(stderr, "### Error: Illegal state encountered:\n");
          for(k=0; k<ysize; k++)
             fprintf(stderr, "\t%lx\n", state[k]);
         }
       }
      count++;
      mask = mask>>1;
     }
   }
#else
  index = 0;
  np = n_part;
  for(j=0; j<ysize; j++)
   {
    lnp = np-((int) np_tab[state[j]]);
    if(lnp < 0)
     {
      fprintf(stderr, "### Error: Illegal state encountered:\n");
      for(k=0; k<ysize; k++)
         fprintf(stderr, "\t%lx\n", state[k]);
     }
    else
     index += index_tab[state[j]][j][lnp];
    np = lnp;
   }
#endif

#ifdef DEBUG
  fprintf(stderr, "State:\n");
  for(k=0; k<ysize; k++)
     fprintf(stderr, "\t%lx\n", state[k]);
  fprintf(stderr, "Index:%ld\n", index);
#endif

  return(index);
 }

/* Translate a state bitwise left */

static void translate_left(unsigned int *state)
 {
  register unsigned int mask;
  register int i;

  mask = ((unsigned int) (1<<xsize)) - 1;
  if(! mask)
    mask = -1;
  for(i=0; i<ysize; i++)
    state[i] = (((state[i]<<1) | ((state[i] >> (xsize - 1)) & 1)
               ) & mask);
 }

/* Translate a state in y-direction */

static void translate_y(unsigned int *state)
 {
  register int k;
  unsigned int temp;

  temp = state[0];
  for(k=0; k<ysize-1; k++)
    state[k] = state[k+1];
  state[ysize-1] = temp;
 }

/* Reflect state along the diagonal of the triangular lattice */

static void reflect_diag(unsigned int *state)
 {
  register int j;
  unsigned int tempa, tempb;

  for(j=0; j<ysize; j++)
   {
     tempa = state[j];
     tempb = tempa & 273;		/* 1, 5 and 9 are invariant */
     if(tempa & 2)			/* 2 <-> 4 */
       tempb |= 8;
     if(tempa & 8)
       tempb |= 2;
     if(tempa & 4)			/* 3 <-> 7 */
       tempb |= 64;
     if(tempa & 64)
       tempb |= 4;
     if(tempa & 32)			/* 6 <-> 8 */
       tempb |= 128;
     if(tempa & 128)
       tempb |= 32;
     state[j] = tempb;
   }
 }

/* Reflect state along the y-direction */

static void reflect_y(unsigned int *state)
 {
  register int j;
  unsigned int temp;

  for(j=0; j<ysize/2; j++)
   {
     temp = state[j];
     state[j] = state[ysize-j-1];
     state[ysize-j-1] = temp;
   }
 }

/* Create the momentum basis */

static void create_momenta()
 {
  register long i;
  register int j, k;
  int flag;
  int tcnt;
  unsigned int *state;
  int refl;
  int sign, fac;
  long index;
  long ps[3000];
  int pcnt;

  charge_momentum = (long *)m_alloc(sizeof(long)*nstates, "create_momenta");
  charge_prefac = (int *)m_alloc(sizeof(int)*nstates, "create_momenta");

  for(i=0; i<nstates; i++)
   {
    charge_momentum[i] = MAGIC;
    charge_prefac[i] = 0;
   }

  pstates = 0;
  for(i=0; i<nstates; i++)
   if(charge_momentum[i] == MAGIC)
    {
     state = index_to_state(i);
     sign = 1;
     pcnt = 0;

     for(refl=0; refl<4; refl++)
      {				/* Orbit over reflection group */
       for(tcnt=0; tcnt<ysize; tcnt++)
        {			/* Orbit over translational group */
         index = state_to_index(state);
         charge_momentum[index] = pstates;   
         charge_prefac[index] += sign;

         for(k=0; (k<pcnt) && (ps[k] != index); k++)
                    ;		/* Remember all states which appeared */
         if(k==pcnt)
            ps[pcnt++] = index;

         translate_y(state);
         sign *= momentum_y;
        }			/* End of translation group */
       if(refl & 1)
        {
         reflect_y(state);
         sign *= parity_y;
        }
       else
        {
         reflect_diag(state);
         sign *= parity_diag;
        }
      }				/* End of reflection group */
     free(state);

     fac = -1;			/* Test prefactors */
     for(j=0; j<pcnt; j++)
        {
         sign = charge_prefac[ps[j]];
         if(sign < 0)
           sign = -sign;
         if(sign)
           charge_prefac[ps[j]] /= sign;
         if(fac < 0)
           fac = sign;
         if(fac != sign)
          {
           fprintf(stderr, "### Error: Not all prefactors have equal magnitude\n");
           exit(-15);
          }
        }

     if(! fac)
      {
       for(j=0; j<pcnt; j++)
           charge_momentum[ps[j]] = MAGIX;
      }
     else
       pstates++;
    }

  momentum_charge = (long *)m_alloc(sizeof(long)*pstates, "create_momenta");
  momentum_prefac = (int *)m_alloc(sizeof(int)*pstates, "create_momenta");

  for(i=0; i < pstates; i++)
   {
    momentum_charge[i] = MAGIC;
    momentum_prefac[i] = 0; 
   }

  for(i=0; i<nstates; i++)
   if(charge_momentum[i] >= 0)
     {
       if(momentum_charge[charge_momentum[i]] < 0)
         momentum_charge[charge_momentum[i]] = i;
       momentum_prefac[charge_momentum[i]]++;
     }
 }

/* Apply Hamilton operator */

void apply_internal_ham(struct vector *src, struct vector *dst, char *name)
 {
  register long i, j, dind;
  register int x;
  int nj;
  unsigned int mask, lmask, rmask;
  unsigned int tempa, tempb;
  long index;
  int szsz, szszp;
  unsigned int *state;
  REAL prefac;

  if(src->el_cnt != pstates)
   {
    display_error("Dimension of input vector doesn't match - can't do internal matrix application");
    return;
   }

  if(dst->el_cnt != pstates)
   {
    display_error("Dimension of output vector doesn't match - can't do internal matrix application");
    return;
   }

  for(i=0; i<dst->el_cnt; i++)		/* First set destination to zero */
   {
     dst->elements[i].im = dst->elements[i].re = 0.0;
   }

  for(i=0; i<pstates; i++)		/* Loop over all momentum states */
   {
    szszp = szsz = 0;
    state = index_to_state(momentum_charge[i]);
    for(j=0; j<ysize; j++)		/* Loop over all sites */
     {
      for(x=0; x<NNEIGH; x++)
       {
        lmask = neighbours[x].lm;	/* triangular plane */
        rmask = neighbours[x].rm;
        tempa = state[j];
        if( ((tempa & lmask) && (! (tempa & rmask))) ||
            ((! (tempa & lmask)) && (tempa & rmask)) )
         {				/* S^+ S^- - interaction */
           state[j] ^= (lmask | rmask);
           dind = charge_momentum[index = state_to_index(state)];
           if(dind >= 0)
            {
             prefac = (jpval/((double) 2));
             if(charge_prefac[momentum_charge[i]] != charge_prefac[index])
                prefac = -prefac;
             if(momentum_prefac[dind] != momentum_prefac[i])
                prefac *= sqrt(((REAL) momentum_prefac[i]) /
                               ((REAL) momentum_prefac[dind]));
             dst->elements[dind].re += (prefac * src->elements[i].re);
             dst->elements[dind].im += (prefac * src->elements[i].im);
            }
           state[j] = tempa;
           szszp--;
         }
        else
           szszp++;
       }

      for(x=0; x<xsize; x++)
       {
        mask = (1 << x);		/* y-direction */ 
        
        nj = (j+1) % ysize;
        tempa = state[j];
        tempb = state[nj];
        if((tempa & mask) != (tempb & mask))
         {				/* S^+ S^- - interaction */
          state[j] ^= mask;
          state[nj] ^= mask;
          dind = charge_momentum[index = state_to_index(state)];
          if(dind >= 0)
            {
             prefac = (jval/((double) 2));
             if(charge_prefac[momentum_charge[i]] != charge_prefac[index])
                prefac = -prefac;
             if(momentum_prefac[dind] != momentum_prefac[i])
                prefac *= sqrt(((REAL) momentum_prefac[i]) /
                              ((REAL) momentum_prefac[dind]));
             dst->elements[dind].re += (prefac * src->elements[i].re);
             dst->elements[dind].im += (prefac * src->elements[i].im);
            }
          state[j] = tempa;
          state[nj] = tempb;
          szsz--;
         }
        else
          szsz++;
       }
     }
    free(state);

    prefac = DELTA/((double) 4)*jval*((REAL) szsz)
             + DELTAP/((double) 4)*jpval*((REAL) szszp);
						/* The diagonal element */
    dst->elements[i].re += (prefac * src->elements[i].re);
    dst->elements[i].im += (prefac * src->elements[i].im);
   }
 }

/* Read parameters */

void read_params()
 {
  FILE *fpi;
  char temp_str[256];
  register int i;
  int line;

  fpi  = fopen("internal-params", "r");
  if(! fpi)
   {
    fprintf(stderr, "### Error: Failed to open `internal-params' for reading\n");
    exit(-14);
   }
  line = 0;
  while(fgets(temp_str, 255, fpi))
   {
    line ++;
    for(i=0; temp_str[i] && (temp_str[i] <= ' ') && (i < 255); i++)
          ;
    if(temp_str[i] && (temp_str[i] != '#') && (temp_str[i] != '%'))
     {
      if(temp_str[i+1] != '=')
        {
         fprintf(stderr, "### Error: Not a `=' as second character on line %d in file `internal-params'\n", line);
         exit(-14);
         }
       switch(temp_str[i] | 0x20)
         {
           case 'l':
             sscanf(temp_str+i+2, "%d", &ysize);
             break;
           case 'n':
             sscanf(temp_str+i+2, "%d", &n_part);
             break;
           case 'k':
             momentum_y = (temp_str[i+2] == '-')?-1:1;
             break;
           case 'r':
             parity_diag = (temp_str[i+2] == '-')?-1:1;
             break;
           case 's':
             parity_y = (temp_str[i+2] == '-')?-1:1;
             break;
           case 'j':
             sscanf(temp_str+i+2, "%lg", &jval);
             break;
           case 'a':
             sscanf(temp_str+i+2, "%lg", &jpval);
             break;
           default:
             temp_str[i+1] = 0;
             fprintf(stderr, "### Error: Unknown `%s' as first character on line %d in file `internal-params'\n", temp_str+i, line);
             exit(-14);
         }
     }
   }
  fclose(fpi);

  if((ysize <= 0) || (ysize > MAXD))
   {
    fprintf(stderr, "### Error: y-size is not between 0 and %d\n", MAXD);
    exit(-14);
   }

  if((n_part < 0) || (n_part > xsize * ysize))
   {
    fprintf(stderr, "### Error: Number of particles is not between 0 and %d\n", xsize * ysize);
    exit(-14);
   }

  fprintf(stderr, "\nL: %d\nnumber of particles: %d\n", ysize, n_part);

  fprintf(stderr, "J = %1.12lg\n", jval);
  fprintf(stderr, "J' = %1.12lg\n", jpval);
  fprintf(stderr, "momentum y: %s\n", (momentum_y == -1)?"-":"+");
  fprintf(stderr, "diagonal parity: %s\n", (parity_diag == -1)?"-":"+");
  fprintf(stderr, "y-parity: %s\n", (parity_y == -1)?"-":"+");
 }

/* Init the internal Hamiltonian application */

long init_interal_h()
 {
  register int i;

  fprintf(stderr, "Internal application of the Hamilton-Matrix\n");
  fprintf(stderr, "for a spin-1/2 Heisenberg model for CsCuCl_3\n");
  fprintf(stderr, "on a 3 x 3 x L lattice\n");
  fprintf(stderr, "written December 1998-March 1999 by Andreas Honecker\n");

  read_params();

  for(i=0; i<NNEIGH; i++)
   {
    neighbours[i].lm = 1 << (neighbours[i].l - 1);
    neighbours[i].rm = 1 << (neighbours[i].r - 1);
   }

  sites = ysize * xsize;	/* Compute the volume of the system */

#ifndef DIRECTBIN
  init_binomials();
#endif
#ifndef SLOWINDEX
  init_mtables();
#endif

  gen_states();

  fprintf(stderr, "\nNumber of states with %d down spins on a 3 x 3 x %d lattice: %ld\n",
                   n_part, ysize, nstates); 

  create_momenta();

  fprintf(stderr, "\nNumber of associated momentum eigenstates: %ld\n",
                   pstates);

  return(pstates);
 }

/* Exit the internal Hamiltonian application */

void exit_interal_h()
 {
#ifndef SLOWINDEX
  free_mtables();
#endif
#ifndef DIRECTBIN
  free_binomials();
#endif

  free(momentum_prefac);		/* Now free the tables used */
  free(momentum_charge);
  free(charge_momentum);
  
  free_states();
 }
