/* Standard-Definitions */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <limits.h>
#include <float.h>
/* Definitions for that program */
#include "useful.h"
#include "complex.h"
#include "matrix.h"
#include "vectors.h"
#include "apply_h.h"

/* The parameters */

#define DELTA 1.0L
double j1val = 1.0L;
double j2val = 1.0L;
double jpval = 1.0L;

/* Two special numbers (should be negative and different, otherwise arbitrary) */

#define MAGIC -12345
#define MAGIX -87654

/* The maximal linear extent */

#define MAXD 32

/* The linear size */

static int size=MAGIC;

/* The number of particles (corresponding to total S^z) */

static int n_part=MAGIC;

/* The variours parities */

static int momentum=1;		/* momentum */
static int parity=1;		/* reflection */

/* Flag if to symmetrize wrt. quadrumers */

static int sym_flag = 0;

/* The number of states in the total-S^z vector space */

static long nstates;

/* Dimension of the momentum eigenspace */

static long pstates;

/* Mapping tables */

static long *charge_momentum;
	/* Which S^z eigenvector belongs to which momentum eigenvector ? */

static signed char *charge_prefac;
	/* The (sign) prefactors */

static long *momentum_charge;
	/* Representative S^z eigenvector for the momentum eigenvector */

static int *momentum_prefac;
	/* The number of states associated to the momentum eigenvector */

/* Mapping tables */

#ifndef SLOWINDEX
long ***index_tab;
char *np_tab;

int tab_size;

/* The maximal number of bits used in these tables */

static long maxbits;

/* The corresponding mask */

static unsigned long mapmask;

/* In how many slices do we split a state for the mapping  ? */

static int map_split;

#endif

/* Table of binomial coefficients */

static long *bin_coeff[1200];

/* ****** Some math routines ****** */

/* define DIRECTBIN in order not to store, but to compute the binomial
   coefficients - useful for very large systems at small densities  */

#ifdef DIRECTBIN
  #define over(a,b) binomial(a,b)
#else
  #define over(a,b) bin_coeff[a][b]
#endif

/* Den groessten gemeinsamen Teiler zweier Zahlen feststellen
   ueber Euklidischen Algorithmus
   Zehnerpotenzen schneller als Primfaktorzerlegung */

static long get_gcd(long a, long b)
 {
   register int i;
   long remainder;

   if(a == 0)           /* Sonderbehandlung fuer Nullen */
     return(1);
   if(b == 0)
     return(1);

   if(a < 0)            /* Betrag bilden */
     a=-a;
   if(b < 0)
     b=-b;

   if(a == 1)           /* Sonderbehandlung fuer Einsen */
     return(1);
   if(b == 1)
     return(1);

   remainder = a % b;   /* Das ist der Euklische Algorithmus */
   while(remainder)
    {
      a=b;
      b=remainder;
      remainder = a % b;
    }           /* Nun steht der GGT in b */

   return(b);
 }

/* Der Binomialkoeffizient */

static long binomial(long n, long m)
 {
   register long i;
   register long gcd;
   register long res1, res2;

   if(m > n-m)                  /* Symmetrie ausnutzen */
     m = n-m;

   if(m < 0)
     return(0);

   if((n==30) && (m==15))
     return(155117520L);

   if((n==31) && (m==13))
     return(206253075L);
   if((n==31) && (m==14))
     return(265182525L);
   if((n==31) && (m==15))
     return(300540195L);

   if((n==32) && (m==12))
     return(225792840L);
   if((n==32) && (m==13))
     return(347373600L);
   if((n==32) && (m==14))
     return(471435600L);
   if((n==32) && (m==15))
     return(565722720L);
   if((n==32) && (m==16))
     return(601080390L);

   res1 = 1;
   res2 = 1;
   for(i=0; i<m; i++)
    {
      res1 *= (n-i);           /* Multiplizieren */
      res2 *= (i+1);
      gcd = get_gcd(res1, res2);      /* Und kuerzen */
      res1 /= gcd;
      res2 /= gcd;
    }
   if(res2 != 1)
     fprintf(stderr, "*** Error: Binomial coefficient (%ld,%ld) not an integer !\n", n, m);
   return(res1);
 }

/*

static long binomial(long n, long m)
 {
  long bin;

  bin = binomial_c(n, m);
  printf("binomial(%ld,%ld) = %ld\n", n, m, bin);
  return(bin);
 }

*/

/* Allocate memory and produce error message if not available */

static void *m_alloc(unsigned long size, char *str)
 {
  void *ptr;

  if(size > 1000L)
    fprintf(stderr, "Allocating %lu bytes from %s()\n", size, str);

  ptr = (void *)malloc(size);

  if(! ptr)
   {
    fprintf(stderr, "### Error: Out of memory in %s()\n", str);
    exit(-13);
   }
  return(ptr);
 }

/* Initialize binomial coefficients */

static void init_binomials()
 {
  register long n, m;

  for(n=0; n<=size; n++)
   {
    bin_coeff[n] = (long *)m_alloc((((n+1)<n_part)?(n+2):(n_part+1))*sizeof(long), "init_binomials");
    for(m=0; m<=n+1; m++)
     if(m <= n_part)
       bin_coeff[n][m] = binomial(n,m);
   }
 }

/* Free binomial coefficients */

static void free_binomials()
 {
  register long n;

  for(n=0; n<=size; n++)
    free(bin_coeff[n]);
 }

#ifndef SLOWINDEX

static char IMSTR[] = "init_mtables";

/* Initialize mapping tables */

static void init_mtables()
 {
  register int i, j, x;
  long mask, index;
  int cnt, np, lnp;

  if(size > 16)
   {
    map_split = 2;
    maxbits = size/2;
   }
  else
   {
    map_split = 1;
    maxbits = size;
   }
  mapmask = (1<<maxbits)-1;

  tab_size = 1<<maxbits;

  np_tab = (char *)m_alloc(tab_size * sizeof(char), IMSTR);

  for(i=0; i<tab_size; i++)
   {
    cnt = 0;
    mask = 1;
    for(x = 0; x<maxbits; x++)
     {
       if(i & mask)
         cnt++;
       mask = mask<<1;
     }
    np = np_tab[i] = cnt;
   }

  index_tab = (long ***)m_alloc(tab_size * sizeof(long **), IMSTR);
  for(i=0; i<tab_size; i++)
   {
    index_tab[i] = (long **)m_alloc(map_split * sizeof(long *), IMSTR);
    for(j=0; j<map_split; j++)
     {
      if(n_part - np_tab[i] >= 0)
        index_tab[i][j] = (long *)m_alloc((n_part - np_tab[i] + 1)
                  * sizeof(long), IMSTR);
      else
        index_tab[i][j] = 0;
     }
   }

  for(i=0; i<tab_size; i++)
   for(j=0; j<map_split; j++)
    for(lnp=0; lnp<=n_part - np_tab[i]; lnp++)
     {
      np = lnp + np_tab[i];
      cnt = j*maxbits;
      index = 0;
      mask = 1<<(maxbits-1);
      for(x=0; x<maxbits; x++)
       {
        if(i & mask)
         {
          index += over(size-cnt-1, np);
          np--;
          if(np < 0)
           fprintf(stderr, "### Error: Illegal bit pattern %x encountered:\n", i);
         }
        cnt++;
        mask = mask>>1;
       }
      index_tab[i][j][lnp] = index; 
     }
 }

/* Free mapping tables */

static void free_mtables()
 {
  register int i, j;

  for(i=0; i<tab_size; i++)
   {
    for(j=0; j<map_split; j++)
       if(index_tab[i][j])
           free(index_tab[i][j]);
    free(index_tab[i]);
   }
  free(index_tab);
  free(np_tab);
 }
#endif

/* Generate states in S^z-symmetry subspace */

static void gen_states()
 {
  nstates = over(size, n_part);
 }

/* Free states in S^z-symmetry subspace */

static void free_states()
 {
 }

/* Convert index to bit representation of state */

static unsigned long index_to_state(long index)
 {
  long oindex;
  unsigned long state;
  unsigned long mask;
  int np, hh;
  int count;
  register int i, j, k;

  state = 0;

  oindex = index;

  count = 0;
  np = n_part;

  mask = 1<<(size-1);
  for(i=0; i<size; i++)
   {
    hh = over(size-count-1, np);
#ifdef DEBUG
    printf("(%d,%d) = %d\n", size-count-1, np, hh);
#endif
    if(index >= hh)
     {
      index -= hh;
      np--;
      state |= mask;
     }
    count++;
    mask = mask>>1;
   }

  if(index || (np < 0))
   fprintf(stderr, "*** Error: failed to translate index `%ld' to state\n", oindex);

#ifdef DEBUG
  fprintf(stderr, "Result of index_to_state %lx:\n", state);
#endif

  return(state);
 }

/* Convert bit representation of state to index of matrix */

static long state_to_index(unsigned long state)
 {
  long index;
#ifdef SLOWINDEX
  unsigned long mask;
  int count;
#else
  unsigned long cstate;
  int lnp;
#endif
  int np;
  register int i, j, k;

#ifdef SLOWINDEX
  count = index = 0;
  np = n_part;
  mask = 1<<(size-1);
  for(i=0; i<size; i++)
   {
    if(state & mask)
     {
      index += over(size-count-1, np);
      np--;
      if(np < 0)
       {
        fprintf(stderr, "### Error: Illegal state encountered: %lx\n", state);
       }
     }
    count++;
    mask = mask>>1;
   }
#else
  index = 0;
  np = n_part;
  for(j=0; j<map_split; j++)
   {
    if(j==(map_split-1))
      cstate = state & mapmask;
    else
      cstate = (state>>maxbits) & mapmask;
    lnp = np-np_tab[cstate];
    if(lnp < 0)
     {
      fprintf(stderr, "### Error: Illegal state encountered %lx:\n", state);
     }
    else
      index += index_tab[cstate][j][lnp];
    np = lnp;
   }
#endif

#ifdef DEBUG
  fprintf(stderr, "State: %lx\n", state);
  fprintf(stderr, "Index: %ld\n", index);
#endif

  return(index);
 }

/* Translate a state bitwise left */

static unsigned long translate_left(unsigned long state)
 {
  register unsigned long mask;

  mask = ((unsigned long) (1<<size)) - 1;
  if(! mask)
    mask = -1;
  state = (((state<<1) | ((state >> (size - 1)) & 1)) & mask);
  return(state);
 }

/* Symmetrize quadrumers in state */

static unsigned long reflect_quadrumer(unsigned long state, int x)
 {
  unsigned long lmask, rmask;
  unsigned long temp;

  lmask = 1<<(4*x+1);
  rmask = 1<<(4*x+2);
  if(((! (state & lmask)) && (state & rmask)) || ((state & lmask) && (! (state & rmask))))
     state ^= (lmask | rmask);
  return(state);
 }

/* Reflect state */

static unsigned long reflect(unsigned long state)
 {
  register int j, x;
  unsigned long temp;
  unsigned long mask, rmask;

  temp = 0;
  rmask = 8;
  mask = 1;
  for(x = 0; x<size; x++)
   {
    if(state & mask)
      temp |= rmask;
    mask = (mask << 1);
    if(rmask == 1)
      rmask = 1 << (size-1);
    else
      rmask = (rmask >> 1);
   }
  return(temp);
 }

/* Create the momentum basis */

#define ORBITLEN 10000

static void create_momenta()
 {
  register long i;
  register int j, k;
  int flag;
  int tcnt;
  unsigned long state;
  int refl;
  int sign, fac;
  long index;
  unsigned long p_states[ORBITLEN];
  int p_prefac[ORBITLEN];
  int pcnt;
  int sym_cnt, lsym_cnt;
  int perc, lastperc;

  fprintf(stderr, "Generating states in symmetry subspace\n"); 

  charge_momentum = (long *)m_alloc(sizeof(long)*nstates, "create_momenta");
  charge_prefac = (signed char *)m_alloc(sizeof(signed char)*nstates, "create_momenta");

  for(i=0; i<nstates; i++)
   {
    charge_momentum[i] = MAGIC;
    charge_prefac[i] = 0;
   }

  lastperc = -1;
  pstates = 0;
  for(i=0; i<nstates; i++)
   {
   perc = (((long long) i) * 100L)/nstates;
   if(lastperc != perc)
     fprintf(stderr, "%d %%\n", perc);
   lastperc = perc;
   if(charge_momentum[i] == MAGIC)
    {
     state = index_to_state(i);
     sign = 1;
     pcnt = 0;

     sym_cnt = 0;
/*
printf("%lx -> ", state);
*/
     while(sym_cnt < ((sym_flag)?(1<<(size/4)):1))
      {				/* Quadrumer symmetrization loop */
     for(refl=0; refl<=1; refl++)
      {				/* Orbit over reflection group */
       for(j=0; j<(size/4); j++)
        {			/* Orbit over translational group */
/*
         index = state_to_index(state);
         charge_momentum[index] = pstates;   
         charge_prefac[index] += sign;
*/

         for(k=0; (k<pcnt) && (p_states[k] != state); k++)
                  ;		/* Remember all states which appeared */
         if(k==pcnt)
           {
            p_prefac[k] = sign;
            p_states[pcnt++] = state;
           }
         else
            p_prefac[k] += sign;
         state=translate_left(state);
         state=translate_left(state);
         state=translate_left(state);
         state=translate_left(state);
         sign *= momentum;
        }			/* End of translation group */
       state = reflect(state);
       sign *= parity;
      }				/* End of reflection group */
       lsym_cnt = sym_cnt;
       sym_cnt++;
       if(sym_flag)
        for(j=0; j<(size/4); j++)
          if((lsym_cnt ^ sym_cnt) & (1<<j))
            state = reflect_quadrumer(state, j);
/*
printf("%lx -> ", state);
*/
      }				/* End of quadrumer symmetrization loop */
/*
printf("\n");
*/

     fac = -1;			/* Test prefactors */
     for(j=0; j<pcnt; j++)
        {
         sign = p_prefac[j];
         if(sign < 0)
           sign = -sign;
         if(sign)
           p_prefac[j] /= sign;
         if(fac < 0)
           fac = sign;
         if(fac != sign)
          {
           fprintf(stderr, "### Error: Not all prefactors have equal magnitude\n");
           exit(-15);
          }
        }

/*
printf("Length of this orbit: %d -- coefficient: %d\n", pcnt, fac);
*/

     for(j=0; j<pcnt; j++)
      {
       index = state_to_index(p_states[j]);
       charge_prefac[index] = p_prefac[j];
       charge_momentum[index] = fac?pstates:MAGIX;
      }
     if(fac)
       pstates++;
    }
   }
  fprintf(stderr, "... now the remaining mapping tables\n");

  momentum_charge = (long *)m_alloc(sizeof(long)*pstates, "create_momenta");
  momentum_prefac = (int *)m_alloc(sizeof(int)*pstates, "create_momenta");

  for(i=0; i < pstates; i++)
   {
    momentum_charge[i] = MAGIC;
    momentum_prefac[i] = 0; 
   }

  for(i=0; i<nstates; i++)
   if(charge_momentum[i] >= 0)
     {
       if(momentum_charge[charge_momentum[i]] < 0)
         momentum_charge[charge_momentum[i]] = i;
       momentum_prefac[charge_momentum[i]]++;
     }

  fprintf(stderr, "... done.\n");
 }

/* Apply Hamilton operator */

void apply_internal_ham(struct vector *src, struct vector *dst, char *name)
 {
  register long i, j, dind;
  register int x;
  int nj;
  unsigned long mask, nmask, nnmask;
  unsigned long tempa, tempb;
  long index;
  int szsz1, szsz2, szszp;
  unsigned long state;
  REAL prefac;

  if(src->el_cnt != pstates)
   {
    display_error("Dimension of input vector doesn't match - can't do internal matrix application");
    return;
   }

  if(dst->el_cnt != pstates)
   {
    display_error("Dimension of output vector doesn't match - can't do internal matrix application");
    return;
   }

  for(i=0; i<dst->el_cnt; i++)		/* First set destination to zero */
   {
     dst->elements[i].im = dst->elements[i].re = 0.0;
   }

  for(i=0; i<pstates; i++)		/* Loop over all momentum states */
   {
    szsz1 = szsz2 = szszp = 0;
    state = index_to_state(momentum_charge[i]);

    mask = 1;
    for(x=0; x<size; x++)		/* Loop over all sites */
     {
      nmask = mask<<1;			/* Nearest neighbour interaction */
      if(x == (size-1))
        nmask = 1;
      if(! (x & 1))
       {
      tempa = state;
      if( ((tempa & mask) && (! (tempa & nmask))) ||
          ((! (tempa & mask)) && (tempa & nmask)) )
       {				/* S^+ S^- - interaction */
         state ^= (mask | nmask);
         dind = charge_momentum[index = state_to_index(state)];
         if(dind >= 0)
          {
           prefac = (jpval/2.0L);
           if(charge_prefac[momentum_charge[i]] != charge_prefac[index])
              prefac = -prefac;
           if(momentum_prefac[dind] != momentum_prefac[i])
              prefac *= sqrt(((REAL) momentum_prefac[i]) /
                             ((REAL) momentum_prefac[dind]));
           dst->elements[dind].re += (prefac * src->elements[i].re);
           dst->elements[dind].im += (prefac * src->elements[i].im);
          }
         state = tempa;
         szszp--;
       }
      else
       {
         szszp++;
       }

       }				/* End of nearest neighbour interaction */

      if(! (x & 2))
       {
      nnmask = nmask<<1;		/* Next nearest neighbour interaction */
      if(x == (size-2))
        nnmask = 1;
      tempa = state;
      if( ((tempa & mask) && (! (tempa & nnmask))) ||
          ((! (tempa & mask)) && (tempa & nnmask)) )
       {                                /* S^+ S^- - interaction */
         state ^= (mask | nnmask);
         dind = charge_momentum[index = state_to_index(state)];
         if(dind >= 0)
          {
           prefac = (jpval/2.0L);
           if(charge_prefac[momentum_charge[i]] != charge_prefac[index])
              prefac = -prefac;
           if(momentum_prefac[dind] != momentum_prefac[i])
              prefac *= sqrt(((REAL) momentum_prefac[i]) /
                             ((REAL) momentum_prefac[dind]));
           dst->elements[dind].re += (prefac * src->elements[i].re);
           dst->elements[dind].im += (prefac * src->elements[i].im);
          }
         state = tempa;
         szszp--;
       }
      else
         szszp++;

       }				/* End of next nearest neighbour interaction */

      /* *** The middle chain *** */

      if(((x & 3) == 0) || ((x & 3) == 3))
       {
        if((x & 3) == 0)
         {
          nnmask = mask<<3;			/* Third neighbour interaction */
          if(x == (size-3))
            nnmask = 1;
         }
        else
         {
          nnmask = mask<<1;			/* Nearest neighbour interaction */
          if(x == (size-1))
            nnmask = 1;
         }
      tempa = state;
      if( ((tempa & mask) && (! (tempa & nnmask))) ||
          ((! (tempa & mask)) && (tempa & nnmask)) )
       {                                /* S^+ S^- - interaction */
         state ^= (mask | nnmask);
         dind = charge_momentum[index = state_to_index(state)];
         if(dind >= 0)
          {
           if((x & 3) == 0)
             prefac = (j2val/2.0L);
           else
             prefac = (j1val/2.0L);
           if(charge_prefac[momentum_charge[i]] != charge_prefac[index])
              prefac = -prefac;
           if(momentum_prefac[dind] != momentum_prefac[i])
              prefac *= sqrt(((REAL) momentum_prefac[i]) /
                             ((REAL) momentum_prefac[dind]));
           dst->elements[dind].re += (prefac * src->elements[i].re);
           dst->elements[dind].im += (prefac * src->elements[i].im);
          }
         state = tempa;
         if((x & 3) == 0)
            szsz2--;
         else
            szsz1--;
       }
      else
       {
         if((x & 3) == 0)
            szsz2++;
         else
            szsz1++;
       }

       }				/* End of middle chain */
        
      mask = nmask;
     }

    prefac = DELTA/4.0L*(j1val*((REAL) szsz1) + j2val*((REAL) szsz2) + jpval*((REAL) szszp));	/* The diagonal element */
    dst->elements[i].re += (prefac * src->elements[i].re);
    dst->elements[i].im += (prefac * src->elements[i].im);
   }
 }

/* Routine to compute correlation functions with respect to the vector in src
   (put an empty one, if you don't need it) */

void compute_correl(struct vector *src, FILE *fpo)
 {
 }

/* Read parameters */

void read_params()
 {
  FILE *fpi;
  char temp_str[256];
  register int i;
  int line;

  fpi  = fopen("internal-params", "r");
  if(! fpi)
   {
    fprintf(stderr, "### Error: Failed to open `internal-params' for reading\n");
    exit(-14);
   }
  line = 0;
  while(fgets(temp_str, 255, fpi))
   {
    line ++;
    for(i=0; temp_str[i] && (temp_str[i] <= ' ') && (i < 255); i++)
          ;
    if((temp_str[i] | 0x20) == 's')
       sym_flag = -1;
    if(temp_str[i] && (temp_str[i] != '#') && (temp_str[i] != '%') && ((temp_str[i] | 0x20) != 's'))
     {
       if(temp_str[i+((temp_str[i] | 0x20) == 'j')?2:1] != '=')
         {
         fprintf(stderr, "### Error: Not a `=' as %s character on line %d in file `internal-params'\n", ((temp_str[i] | 0x20) == 'j')?"third":"second", line);
         exit(-14);
         }
       switch(temp_str[i] | 0x20)
         {
           case 'l':
             sscanf(temp_str+i+2, "%d", &size);
             break;
           case 'n':
             sscanf(temp_str+i+2, "%d", &n_part);
             break;
           case 'k':
             momentum = (temp_str[i+2] == '-')?-1:1;
             break;
           case 'p':
             parity = (temp_str[i+2] == '-')?-1:1;
             break;
           case 'j':
             switch(temp_str[i+1])
              {
               case 'p':
                 sscanf(temp_str+i+3, "%lg", &jpval);
                 break;
               case '0':
                 sscanf(temp_str+i+3, "%lg", &j1val);
                 j2val = j1val;
                 break;
               case '1':
                 sscanf(temp_str+i+3, "%lg", &j1val);
                 break;
               case '2':
                 sscanf(temp_str+i+3, "%lg", &j2val);
                 break;
               default:
                 fprintf(stderr, "### Error: Unkown parameter on line %d in file `internal-params'\n", temp_str+i, line);
                 exit(-1414);
              }
             break;
           default:
             temp_str[i+1] = 0;
             fprintf(stderr, "### Error: Unknown `%s' as first character on line %d in file `internal-params'\n", temp_str+i, line);
             exit(-14);
         }
     }
   }
  fclose(fpi);

  if((size <= 0) || (size > MAXD))
   {
    fprintf(stderr, "### Error: size is not between 0 and %d\n", MAXD);
    exit(-14);
   }

  if((n_part < 0) || (n_part > size))
   {
    fprintf(stderr, "### Error: Number of particles is not between 0 and %d\n", size);
    exit(-14);
   }

  if(size & 3)
   {
    fprintf(stderr, "### Error: Size %d is not divisible by 4\n", size);
    exit(-14);
   }

  fprintf(stderr, "\nsize: %d\nnumber of particles: %d\n", size, n_part);

  fprintf(stderr, "J1 = %1.12lg\n", j1val);
  fprintf(stderr, "J2 = %1.12lg\n", j2val);
  fprintf(stderr, "J' = %1.12lg\n", jpval);
  fprintf(stderr, "momentum: %s\n", (momentum == -1)?"-":"+");
  fprintf(stderr, "parity: %s\n", (parity == -1)?"-":"+");
  if(! sym_flag)
    fprintf(stderr, "Do not s");
  else
    fprintf(stderr, "S");
  fprintf(stderr, "ymmetrize the individual quadrumers\n");
 }

/* Init the internal Hamiltonian application */

long init_interal_h()
 {
  register int i;

  fprintf(stderr, "Internal application of the Hamilton-Matrix\n");
  fprintf(stderr, "for a trimerized 1D spin-1/2 Heisenberg model\n");
  fprintf(stderr, "written December 1998-May 2000 by Andreas Honecker\n");

  read_params();

#ifndef DIRECTBIN
  init_binomials();
#endif
#ifndef SLOWINDEX
  init_mtables();
#endif

  gen_states();

  fprintf(stderr, "\nNumber of states with %d down spins on a length %d chain: %ld\n",
                   n_part, size, nstates); 

  create_momenta();

  fprintf(stderr, "\nNumber of associated momentum eigenstates: %ld\n",
                   pstates);

  return(pstates);
 }

/* Exit the internal Hamiltonian application */

void exit_interal_h()
 {
#ifndef SLOWINDEX
  free_mtables();
#endif
#ifndef DIRECTBIN
  free_binomials();
#endif

  free(momentum_prefac);		/* Now free the tables used */
  free(momentum_charge);
  free(charge_momentum);
  
  free_states();
 }
