/* Standard-Definitions */
#include <stdio.h>
/* Definitions for that program */
#include "useful.h"
#include "text.h"
#include "arguments.h"
#include "complex.h"
#include "matrix.h"
#include "vectors.h"

/* ************************************************
   Funktionen zur Argument-Verwaltung */

/* Skip over spaces */

int skip_spaces(char *s, int ofs)
 {
   while(s[ofs] && (s[ofs] <= ' '))
     ofs++;

   return(ofs);
 }

/* Compute minimum/maximum value of a list */
 
static int get_minmax(UBYTE *s, int minmax, int ofs, int *num)
 {
  int temp;

  if(s[ofs] != '{')
    display_error("Argument should start with '{'");
  else
    ofs++;
 
  ofs = get_number(s, ofs, num);

  while(s[ofs] && (s[ofs] != '}') && (s[ofs] != ')'))
    {
     ofs = skip_spaces(s, ofs);		/* Skip spaces */
     if(s[ofs] != ',')
       display_error("Missing komma ','");
     else
       ofs++;
     ofs = get_number(s, ofs, &temp);
     if(minmax > 0)
      {
       if(temp < *num)
         *num = temp;
      }
     else
      {
       if(temp > *num)
         *num = temp;
      }
    }

  if(s[ofs] != '}')
    display_error("Missing '}'");
  else
    ofs++;

  return(ofs);
 }

/* Eine Zahl holen */

int get_number(UBYTE *str, int ofs, int *num)
 {
   int i;
   int sign;
   int psign, result;
   int minmax;
   char fname, name[80];
   struct vector *vptr;
   struct matrix *mptr;

   psign = 1;
   *num = 0;

gloop:
   ofs = skip_spaces(str, ofs);			/* Skip spaces */

   if(! strncasecmp(str+ofs, "dim", 3))		/* Is it a "dim"ension ? */
    {
     ofs += 3;

     result = -1;

     fname = 0;
     if(str[ofs] != '(')
       fname = (str[ofs++] | 0x20);
     if(str[ofs] != '(')
      {
       display_error("No opening bracket after 'dim'");
      }
     ofs++;
     i=0;
     while(str[ofs] && (str[ofs] != ')'))
      {
       name[i++] = str[ofs++];
      }
     name[i] = 0;
     if(str[ofs] != ')')
      {
       display_error("No closing bracket after 'dim'");
      }
     else
      ofs++;
     switch(fname)
       {
         case 0:
          vptr = identify_vector(name, TRUE);
          if(vptr)
            result = vptr->el_cnt;
          break;
         case 'c':
          mptr = identify_matrix(name);
          if(mptr)
            result = mptr->columns;
          break;
         case 'r':
          mptr = identify_matrix(name);
          if(mptr)
            result = mptr->rows;
          break;
         default:
          display_error("Unkown 'dim'ension function");
          result = -1;
          break;
       }
    }
   else
    {
     minmax = 0;
     if(! strncasecmp(str+ofs, "min", 3))	/* Is it a "min" ? */ 
       minmax = 1;
     if(! strncasecmp(str+ofs, "max", 3))	/* or is it a "max" ? */
       minmax = -1;
     if(minmax)
      {
       ofs += 3; 
       if(str[ofs] != '(')
        {
         display_error("Missing opening bracket");
        } 
       else
         ofs++;
       ofs = get_minmax(str, minmax, ofs, &result);
       if(str[ofs] != ')')
        {
         display_error("Missing closing bracket");
        }
       else
         ofs++;
      }
     else
      {
       switch(str[ofs])
         {
           case '-':
             ofs++;
             sign=-1;
             break;
           case '+':
             ofs++;
           default:
             sign=1;
             break;
         }
       ofs = skip_spaces(str, ofs);		/* Skip spaces */ 
       result=0;
       while((str[ofs] >= '0') && (str[ofs] <= '9'))
         {
           result *= 10;
           result += (str[ofs] - '0');
           ofs++;
         }
       result *= sign;
      }
    }

   *num += (result*psign);

   ofs = skip_spaces(str, ofs);			/* Skip spaces */ 

   if(str[ofs] == '+')
    {
     ofs++;
     psign = 1;
     goto gloop;
    }
   if(str[ofs] == '-')
    {
     ofs++;
     psign = -1;
     goto gloop;
    }

   return(ofs);
 }

/* Ein naechsten, durch Leerzeichen getrenntes Argument suchen */

int seek_next_arg(UBYTE *str)
 {
  register int i;

  for(i=0; str[i] > ' '; i++)		/* Find first space or the like */
         ;

  while(str[i] && (str[i] <= ' '))	/* Skip spaces */
      str[i++] = 0;

  return(i);
 }

/* Ein naechstes, durch Kommata abgetrenntes Argument suchen */

UBYTE *get_next_arg(UBYTE *str)
 {
   int bra_level;
   int ofs;

   bra_level = 0;
   for(ofs=0; str[ofs]; ofs++)
    {
      switch(str[ofs])
       {
         case '(':
           bra_level++;
           break;
         case ')':
           bra_level--;
           if(bra_level < 0)
             {
               display_error_text("Too many right brackets in '%s'", str);
               return(0);
             }
           break;
         case ',':
           if(! bra_level)
             {
                str[ofs] = 0;
                return(str+ofs+1);
             }
           break;
       }
    }
   if(bra_level)
      display_error_text("Unmatched left bracket in '%s'", str);
   else
      display_error_text("No komma found in '%s'", str);
   return(0);
 }

/* Locate next argument */

UBYTE *skip_to_next_arg(UBYTE *str)
 {
   int bra_level;
   int ofs;
 
   bra_level = 0;
   for(ofs=0; str[ofs]; ofs++)
    {
      switch(str[ofs])
       {
         case '(':
           bra_level++;
           break;
         case ')':
           bra_level--;
           if(bra_level < 0)
             {
               display_error_text("Too many right brackets in '%s'", str);
               return(0);
             }
           break;
         case ',':
           if(! bra_level)
             {
                str[ofs] = 0;
                return(str+ofs+1);
             }
           break;
       }
    }
   if(bra_level)
      display_error_text("Unmatched left bracket in '%s'", str);
   return(0);
 }
