/* Standard-Definitions */
#include <stdio.h>
#include <math.h>
#include <limits.h>
#include <float.h>
/* Definitions for that program */
#include "useful.h"
#include "complex.h"
#include "variables.h"
#include "text.h"
#include "matrix.h"
#include "vectors.h"
#include "extrapolate.h"
#include "evaluate.h"
#include "functions.h"
#include "integrate.h"
#include "combinat.h"

#include "arguments.h"
#include "help.h"
#include "hamilton.h"
#include "set.h"

#include "date.h"

/* *********************************************
   Routinen um einen Ausdruck zu berechnen */

/* Zugehoerige Schliess-Klammer finden */

int find_close_bra(UBYTE *expression, int start)
 {
   register int i, depth;

   depth = 1;
   for(i=start; expression[i]; i++)
    {
      if(expression[i] == '(')
          depth++;
      if(expression[i] == ')')
       {
          depth--;
          if(! depth)
             return(i);
       }
    }
   display_error_text("Missing right paranthesis in '(%s'", expression + start);
   return(i);
 }

/* Zugehoerige Oeffnungs-Klammer finden */

int find_open_bra(UBYTE *expression, int start)
 {
   register int i, depth;

   depth = 1;
   for(i=start; i >= 0; i--)
    {
      if(expression[i] == ')')
          depth++;
      if(expression[i] == '(')
       {
          depth--;
          if(! depth)
             return(i);
       }
    }
   display_error_text("Missing left paranthesis in '%s)'", expression + start);
   return(i);
 }

/* Funktions-Tokens */

#define F_UNKNOWN 0
#define F_EXP     0x1222
#define F_SIN     0x1223
#define F_COS     0x1224
#define F_SQRT    0x1225
#define F_TAN     0x1226
#define F_COTAN   0x1227
#define F_SINH    0x1229
#define F_COSH    0x1230
#define F_TANH    0x1231
#define F_COTANH  0x1232
#define F_LN      0x1233
#define F_ABS     0x1234
#define F_ARG     0x1235
#define F_ASINH   0x1236
#define F_ACOSH   0x1237
#define F_ATANH   0x1238
#define F_ACOTANH 0x1239
#define F_ASIN    0x1240
#define F_ACOS    0x1241
#define F_ATAN    0x1242
#define F_ACOTAN  0x1243
#define F_RE      0x1244
#define F_IM      0x1245
#define F_CC      0x1246
#define F_BESS_I0	0x1247
#define F_BESS_K0	0x1248

/* Funktion identifizieren */

static int identify_function(UBYTE *str)
 {
   UBYTE s0, s1;

   s0 = (str[0] | 0x20);	/* First character of "str"ing */
   s1 = (str[1] | 0x20);	/* Second character of "str"ing */

   if(s0 == 'e')
    {
     if(tcompare(str+1, (UBYTE *) "xp"))
        return(F_EXP);	/* "exp" */
    }
   if(s0 == 's')
    {
     if(s1 == 'i')
      {
       if(tcompare(str+2, (UBYTE *) "n"))
          return(F_SIN);	/* "sin" */
       if(tcompare(str+2, (UBYTE *) "nh"))
          return(F_SINH);	/* "sinh" */
      }
     if(tcompare(str+1, (UBYTE *) "qrt"))
        return(F_SQRT);	/* "sqrt" */
    }
   if(s0 == 'c')
    {
     if(s1 == 'o')
      {
       if(tcompare(str+2, (UBYTE *) "s"))
          return(F_COS);	/* "cos" */
       if(tcompare(str+2, (UBYTE *) "tan"))
          return(F_COTAN);	/* "cotan" */
       if(tcompare(str+2, (UBYTE *) "sh"))
          return(F_COSH);	/* "cosh" */
       if(tcompare(str+2, (UBYTE *) "tanh"))
          return(F_COTANH);	/* "cotanh" */
      }
     if(tcompare(str+1, (UBYTE *) "c"))
        return(F_CC);	/* "cc" */
    }
   if(s0 == 'a')
    {
     if(s1 == 's')
      {
       if(tcompare(str+2, (UBYTE *) "inh"))
          return(F_ASINH);	/* "asinh" */ 
       if(tcompare(str+2, (UBYTE *) "in"))
          return(F_ASIN);	/* "asin" */
      }
     if(s1 == 'c')
      {
       if(tcompare(str+2, (UBYTE *) "osh"))
          return(F_ACOSH);	/* "acosh" */
       if(tcompare(str+2, (UBYTE *) "otanh"))
          return(F_ACOTANH);	/* "acotanh" */
       if(tcompare(str+2, (UBYTE *) "os"))
          return(F_ACOS);	/* "acos" */
       if(tcompare(str+2, (UBYTE *) "otan"))
          return(F_ACOTAN);	/* acotan" */
      }
     if(s1 == 't')
      {
       if(tcompare(str+2, (UBYTE *) "anh"))
          return(F_ATANH);	/* "atanh" */
       if(tcompare(str+2, (UBYTE *) "an"))
          return(F_ATAN);	/* "atan" */
      }
     if(tcompare(str+1, (UBYTE *) "bs"))
        return(F_ABS);	/* "abs" */
     if(tcompare(str+1, (UBYTE *) "rg"))
        return(F_ARG);	/* "arg" */
    }
   if(s0 == 'b')
    if(s1 == 'e')
     {
     if(compare(str+2, (UBYTE *) "sselI0"))
        return(F_BESS_I0);	/* "BesselI0" */
     if(compare(str+2, (UBYTE *) "sselK0"))
        return(F_BESS_K0);	/* "BesselK0" */
    }
   if(s0 == 't')
    if(s1 == 'a')
     {
      if(tcompare(str+2, (UBYTE *) "n"))
         return(F_TAN);		/* "tan") */
      if(tcompare(str+2, (UBYTE *) "nh"))
         return(F_TANH);	/* "tanh" */
     }

   if(tcompare(str, (UBYTE *) "ln"))
      return(F_LN);
   if(tcompare(str, (UBYTE *) "re"))
      return(F_RE);
   if(tcompare(str, (UBYTE *) "im"))
      return(F_IM);

   return(F_UNKNOWN);
 }

/* Funktions-Tokens fuer Funktionen mit Vektoren */

#define FV_SPROD     0x4321
#define FV_NORM      0x4322
#define FV_NORMALIZE 0x4323
#define FV_APPLYMAT  0x4324
#define FV_EXTVBS    0x4325
#define FV_EXTSTB    0x4326
#define FV_INTPOLY   0x4327
#define FV_ADD       0x4328
#define FV_PROD      0x4329
#define FV_CPROD     0x4330
#define FV_GROUNDST  0x4331
#define FV_FOURIER   0x4332
#define FV_CFOURIER  0x4333
#define FV_LINREG    0x4334
#define FV_ADJOINT   0x4335
#define FV_MPROD     0x4336
#define FV_CGSOLVE   0x4337
#define FV_DIM       0x4338
#define FV_DIMC      0x4339
#define FV_DIMR      0x4340
#define FV_MADD      0x4341
#define FV_COMMUTE   0x4342
#define FV_BCGSOLVE  0x4343

/* Vektor-Funktion identifizieren */

static int identify_vector_function(UBYTE *str)
 {
   UBYTE s0;

   s0 = (str[0] | 0x20);	/* First character of "str"ing */

   if(s0 == 's')
    {
     if(tcompare(str+1, (UBYTE *) "prod"))
        return(FV_SPROD);	/* "sprod" */
     if(tcompare(str+1, (UBYTE *) "mpy"))
        return(FV_PROD);	/* "smpy" */
    }
   if(s0 == 'n')
    if((str[1] | 0x20) == 'o')
    {
     if(tcompare(str+2, (UBYTE *) "rm"))
        return(FV_NORM);	/* "norm" */
     if(tcompare(str+2, (UBYTE *) "rmalize"))
        return(FV_NORMALIZE);	/* "normalize" */
    }
   if(s0 == 'a')
    {
     if(compare(str+1, (UBYTE *) "pply_matrix"))
        return(FV_APPLYMAT);	/* "apply_matrix" */
     if(tcompare(str+1, (UBYTE *) "dd"))
        return(FV_ADD);		/* "add" */
     if(tcompare(str+1, (UBYTE *) "djoint"))
        return(FV_ADJOINT);	/* "adjoint" */
    }
   if(s0 == 'e')
    if((str[1] | 0x20) == 'x')
    {
     if(compare(str+2, (UBYTE *) "trapolate_vbs"))
        return(FV_EXTVBS);	/* "extrapolate_vbs" */
     if(compare(str+2, (UBYTE *) "trapolate_stb"))
        return(FV_EXTSTB);	/* "extrapolate_stb" */
    }
   if(s0 == 'c')
    {
     if(tcompare(str+1, (UBYTE *) "ross"))
        return(FV_CPROD);	/* "cross" */
     if(compare(str+1, (UBYTE *) "omplex_fourier"))
        return(FV_CFOURIER);	/* "complex_fourier" */
     if(tcompare(str+1, (UBYTE *) "ommute"))
        return(FV_COMMUTE);	/* "commute" */
     if(compare(str+1, (UBYTE *) "g_solve"))
        return(FV_CGSOLVE);	/* "cg_solve" */
    }
   if(s0 == 'm')
    {
     if(tcompare(str+1, (UBYTE *) "add"))
        return(FV_MADD);	/* "madd" */
     if(tcompare(str+1, (UBYTE *) "prod"))
        return(FV_MPROD);	/* "mprod" */
    }
   if(s0 == 'd')
    if((str[1] | 0x20) == 'i')
    {
     if(tcompare(str+2, (UBYTE *) "m"))
        return(FV_DIM);	/* "dim" */
     if(tcompare(str+2, (UBYTE *) "mc"))
        return(FV_DIMC);	/* "dimc" */
     if(tcompare(str+2, (UBYTE *) "mr"))
        return(FV_DIMR);	/* "dimr" */
    }

   if(compare(str, (UBYTE *) "bcg_solve"))
      return(FV_BCGSOLVE);
   if(compare(str, (UBYTE *) "interpolate_poly"))
      return(FV_INTPOLY);
   if(tcompare(str, (UBYTE *) "groundstate"))
      return(FV_GROUNDST);
   if(tcompare(str, (UBYTE *) "fourier"))
      return(FV_FOURIER);
   if(compare(str, (UBYTE *) "lin_reg"))
      return(FV_LINREG);

   return(F_UNKNOWN);
 }

/* Funktions-Tokens fuer Funktionen mit mehreren Argumenten */

#define FMARG_INT	0x7f01
#define FMARG_INTOPEN	0x7f02
#define FMARG_INTEXTR	0x7f03
#define FMARG_INTOPENEXTR	0x7f04
#define FMARG_INTCAUCHY	0x7f05
#define FMARG_DIRECDF	0x7f06
#define FMARG_DF	0x7f07
#define FMARG_MIN	0x7f08
#define FMARG_MAX	0x7f09
#define FMARG_BINOMIAL	0x7f0a


/* Identifiziere Funktion mit mehreren Argumenten */

static int identify_marg_function(UBYTE *str)
 {
   UBYTE s0;

   s0 = (str[0] | 0x20);	/* First character of "str"ing */

   if(s0 == 'i')
    if((str[1] | 0x20) == 'n')
     {
      if(tcompare(str+2, (UBYTE *) "t"))
         return(FMARG_INT);	/* "int" */
      if(tcompare(str+2, (UBYTE *) "topen"))
         return(FMARG_INTOPEN);	/* "intopen" */
      if(compare(str+2, (UBYTE *) "t_extrapolate"))
         return(FMARG_INTEXTR);	/* "int_extrapolate" */
      if(compare(str+2, (UBYTE *) "topen_extrapolate"))
         return(FMARG_INTOPENEXTR);	/* "intopen_extrapolate" */
      if(compare(str+2, (UBYTE *) "t_cauchy"))
         return(FMARG_INTCAUCHY);	/* "int_cauchy" */
     }
   if(s0 == 'd')
    {
     if(compare(str+1, (UBYTE *) "irec_df"))
        return(FMARG_DIRECDF);	/* "direc_df" */
     if(tcompare(str+1, (UBYTE *) "f"))
        return(FMARG_DF);	/* "df" */
    }
   if(s0 == 'm')
    {
     if(tcompare(str+1, (UBYTE *) "in"))
        return(FMARG_MIN);	/* "min" */ 
     if(tcompare(str+1, (UBYTE *) "ax"))
        return(FMARG_MAX);	/* "max" */
    }
   if(tcompare(str, (UBYTE *) "binomial"))
      return(FMARG_BINOMIAL);

   return(F_UNKNOWN);
 }

/* Determine smallest or largest value of a list */

void minmaxlist(UBYTE *s, COMPLEX *result, int minmax)
 {
  UBYTE *ns;
  COMPLEX temp;
  register int i;

  if(*s != '{')
    display_error("Argument should start with '{'");
  else
    s++;

  for(i=0; s[i]; i++)
        ;

  if(i)
   {
    if(s[i-1] == '}')
       s[i-1] = 0;
    else
       display_error("Missing '}'");
   }
  else
    display_error("Void argument string");
  

  ns = get_next_arg(s);
  evaluate(s, result);

  while(ns)
    {
     s = ns;
     ns = skip_to_next_arg(s);
     evaluate(s, &temp);
     if(minmax > 0)
      {
       if(temp.re < result->re)
        {
         result->re = temp.re;
         result->im = temp.im;
        }
      }
     else
      {
       if(temp.re > result->re)
        {
         result->re = temp.re;
         result->im = temp.im;
        }
      }
    }
 }


/* *****************************
   Ausdruck berechnen */

void evaluate(UBYTE *string, COMPLEX *result)
 {
  COMPLEX a, b;
  int str_len;
  UBYTE charac;
  UBYTE *expression, *arg1, *arg2, *arg3, *arg4;
  struct variable *ptr;
  int f_id, flag;
  register int i;
  int ofs, ofs1;
  int exponent, t_exponent;
  int neg_flag, p_flag;
  struct vector *vector, *hvector, *hhvector;
  struct matrix *matrix, *hmatrix, *hhmatrix;
  struct function *function;
  UBYTE temp_str[100];

  result->re = 0;
  result->im = 0;

  if(DEBUGFLAG > 1)
    printf("# Evaluating: '%s'\n", string);

  str_len = strlen(string);
  expression=(UBYTE *) malloc(str_len + 1);	/* Eingabe kopieren */
  if(! expression)
    {
      display_error_text("Out of memory during evaluation of: %s", string);
      return;
    }
  for(i=0; expression[i]=string[i]; i++)
         ;


  if(*expression == '(')		/* Klammern aufloesen */
   {
     i = find_close_bra(expression, 1);
     if(! expression[i+1])
      {
        expression[i]=0;
        evaluate(expression + 1, result);
        goto return_it;
      }
   }

  for(i=str_len; i>=0; --i)		/* Addition/Subtraktion durchfuehren */
   {
     if(i > 0)
      {
        flag=TRUE;			/* Achtung bei Zahlen, die e enthalten */
        if(i > 1)
          if((expression[i-2] >= '0') && (expression[i-2] <= '9'))
                 flag=FALSE;
        if((expression[i] == '+') || (expression[i] == '-'))
         if((conv_to_lower(expression[i-1]) != 'e') || flag)
          {
           charac=expression[i];
           expression[i]=0;
           evaluate(expression, &a);
           evaluate(expression+i+1, &b);
           if(charac == '+')
            {
             add(&a, &b, result);
            }
           else
            {
             sub(&a, &b, result);
            }
           goto return_it;
          }
      }
     if(expression[i] == ')')
        i = find_open_bra(expression, i-1);
   }

  if(*expression == '-')		/* Negatives Vorzeichen */
   {
     evaluate(expression+1, result);
     result->re = -result->re;
     result->im = -result->im;
     goto return_it;
   }

  if(*expression == '+')		/* Positives Vorzeichen */
   {
     evaluate(expression+1, result);
     goto return_it;
   }

  for(i=str_len; i>=0; --i)		/* Multiplikation durchfuehren */
   {
     if(i > 0)
      {
        if((expression[i] == '*') && (expression[i-1] != '*') && (expression[i+1] != '*'))
         {
          expression[i]=0;
          evaluate(expression, &a);
          evaluate(expression+i+1, &b);
          multiply(&a, &b, result);
          goto return_it;
         }
      }
     if(expression[i] == ')')
        i = find_open_bra(expression, i-1);
   }

  for(i=str_len; i>=0; --i)		/* Division durchfuehren */
   {
     if(i > 0)
      {
        if(expression[i] == '/')
         {
          expression[i]=0;
          evaluate(expression, &a);
          evaluate(expression+i+1, &b);
          divide(&a, &b, result);
          goto return_it;
         }
      }
     if(expression[i] == ')')
        i = find_open_bra(expression, i-1);
   }

  for(i=str_len; i>=0; --i)		/* Potenzen berechnen */
   {
     if(i > 0)
      {
        if((expression[i] == '*') && (expression[i+1] == '*'))
         {
          expression[i]=0;
          evaluate(expression, &a);
          evaluate(expression+i+2, &b);
          complex_pow(&a, &b, result);
          goto return_it;
         }
        if(expression[i] == '^')	/* Eine Alternative: ^ statt ** */
         {
          expression[i]=0;
          evaluate(expression, &a);
          evaluate(expression+i+1, &b);
          complex_pow(&a, &b, result);
          goto return_it;
         }
      }
     if(expression[i] == ')')
        i = find_open_bra(expression, i-1);
   }

  /* Funktionen */
  charac=conv_to_lower(*expression);
  if((charac >= 'a') && (charac <= 'z'))
   {
     for(i=0; expression[i]; i++)
      if(expression[i] == '(')
       {
        ofs = find_close_bra(expression, i+1);
        if(expression[ofs+1])
         {
           display_error_text("Impromper argument '%s'", expression);
           goto return_it;
         }
        expression[ofs] = 0;
        expression[i] = 0;
        arg1 = expression+i+1;
        f_id = identify_vector_function(expression);	/* Ist es eine Vektor-Funktion ? */
        switch(f_id)
         {
           case FV_SPROD:
             if(arg2=get_next_arg(arg1))
               {
                 vector = identify_vector(arg1, TRUE);
                 hvector = identify_vector(arg2, TRUE);
                 if(vector && hvector)
                   sprod(vector, hvector, result);
                 goto return_it;
               }
             display_error("Improper arguments of sprod");
             result->re = result->im = 0;
             goto return_it;
           case FV_NORM:			/* Norm eines Vektors berechnen */
             vector = identify_vector(arg1, TRUE);
             if(vector)
                result->re = norm(vector);
             result->im = 0;
             goto return_it;
           case FV_NORMALIZE:			/* Vektor auf 1 normieren */
             vector = identify_vector(arg1, TRUE);
             if(vector)
                normalize_vector(vector);
             result->re = result->im = 0;
             goto return_it;
           case FV_APPLYMAT:
             result->re = result->im = 0;
             if(arg2=get_next_arg(arg1))
               if(arg3=get_next_arg(arg2))
                 {
                   matrix = identify_matrix(arg1);
                   vector = identify_vector(arg2, TRUE);
                   hvector = init_dest_vector(arg3, matrix->rows);
                   if(vector && hvector && matrix)
                      apply_named_matrix((char *) arg1, matrix, vector, hvector);
                   goto return_it;
                 }
             display_error("Improper arguments of apply_matrix");
             goto return_it;
           case FV_ADJOINT:
             result->re = result->im = 0;
             if(arg2=get_next_arg(arg1))
               {
                 matrix = identify_matrix(arg1);
                 if(matrix)
                  {
                   alloc_matrix(arg2, matrix->columns, matrix->rows);
                   hmatrix = identify_matrix(arg2);
                   if(hmatrix)
                      adjoint_matrix(matrix, hmatrix);
                  }
                 goto return_it;
               }
             display_error("Improper arguments of adjoint");
             goto return_it;
           case FV_MADD:
             result->re = result->im = 0;
             if(arg2=get_next_arg(arg1))
               if(arg3=get_next_arg(arg2))
                 {
                   matrix = identify_matrix(arg1);
                   hmatrix = identify_matrix(arg2);
                   if(matrix && hmatrix)
                    {
                     alloc_matrix(arg3, matrix->columns, matrix->rows);
                     hhmatrix = identify_matrix(arg3);
                     if(hmatrix)
                        madd(matrix, hmatrix, hhmatrix);
                    }
                   goto return_it;
                 }
             display_error("Improper arguments of madd");
             goto return_it;
           case FV_MPROD:
             result->re = result->im = 0;
             if(arg2=get_next_arg(arg1))
               if(arg3=get_next_arg(arg2))
                 {
                   matrix = identify_matrix(arg1);
                   hmatrix = identify_matrix(arg2);
                   if(matrix && hmatrix)
                    {
                     alloc_matrix(arg3, hmatrix->columns, matrix->rows);
                     hhmatrix = identify_matrix(arg3);
                     if(hmatrix)
                        mprod(matrix, hmatrix, hhmatrix);
                    }
                   goto return_it;
                 }
             display_error("Improper arguments of mprod");
             goto return_it;
           case FV_COMMUTE:
             result->re = result->im = 0;
             if(arg2=get_next_arg(arg1))
               if(arg3=get_next_arg(arg2))
                 {
                   matrix = identify_matrix(arg1);
                   hmatrix = identify_matrix(arg2);
                   if(matrix && hmatrix)
                    {
                     alloc_matrix(arg3, hmatrix->columns, matrix->rows);
                     hhmatrix = identify_matrix(arg3);
                     if(hmatrix)
                        commute(matrix, hmatrix, hhmatrix);
                    }
                   goto return_it;
                 }
             display_error("Improper arguments of commute");
             goto return_it;
           case FV_GROUNDST:
             result->re = result->im = 0;
             if(arg2=get_next_arg(arg1))
               if(arg3=get_next_arg(arg2))
                 if(arg4=get_next_arg(arg3))
                  {
                    matrix = identify_matrix(arg1);
                    vector = init_dest_vector(arg2, matrix->rows);
                    evaluate(arg3, &a);
                    evaluate(arg4, &b);
                    if(matrix && vector)
                         vector_iterate(matrix, vector, &a, &b, result);
                    goto return_it;
                  }
             display_error("Improper arguments of groundstate");
             goto return_it;
           case FV_CGSOLVE:
             result->re = result->im = 0;
             if(arg2=get_next_arg(arg1))
               if(arg3=get_next_arg(arg2))
                 if(arg4=get_next_arg(arg3))
                  {
                    matrix = identify_matrix(arg1);
                    vector = init_dest_vector(arg2, matrix->columns);
                    hvector = identify_vector(arg3, TRUE);
                    evaluate(arg4, &a);
                    if(matrix && vector && hvector)
                         cg_solve(matrix, vector, hvector, &a);
                    goto return_it;
                  }
             display_error("Improper arguments of cg_solve");
             goto return_it;
           case FV_BCGSOLVE:
             result->re = result->im = 0;
             if(arg2=get_next_arg(arg1))
               if(arg3=get_next_arg(arg2))
                 if(arg4=get_next_arg(arg3))
                  {
                    matrix = identify_matrix(arg1);
                    vector = init_dest_vector(arg2, matrix->columns);
                    hvector = identify_vector(arg3, TRUE);
                    evaluate(arg4, &a);
                    if(matrix && vector && hvector)
                         bcg_solve(matrix, vector, hvector, &a);
                    goto return_it;
                  }
             display_error("Improper arguments of bcg_solve");
             goto return_it;
           case FV_ADD:
             result->re = result->im = 0;
             if(arg2=get_next_arg(arg1))
               if(arg3=get_next_arg(arg2))
                 {
                   vector = identify_vector(arg1, TRUE);
                   if(vector)
                    {
                     hvector = identify_vector(arg2, TRUE);
                     hhvector = init_dest_vector(arg3, vector->el_cnt);
                     if(hvector && hhvector)
                        add_vectors(vector, hvector, hhvector);
                     goto return_it;
                    }
                 }
             display_error("Improper arguments of add");
             goto return_it;
           case FV_PROD:
             result->re = result->im = 0;
             if(arg2=get_next_arg(arg1))
               if(arg3=get_next_arg(arg2))
                 {
                    evaluate(arg1, &a);
                    vector = identify_vector(arg2, TRUE);
                    if(vector)
                     {
                        hvector = init_dest_vector(arg3, vector->el_cnt);
                        if(hvector)
                          smpy_vector(&a, vector, hvector);
                     }
                    goto return_it;
                 }
             display_error("Improper arguments of smpy");
             goto return_it;
           case FV_CPROD:
             result->re = result->im = 0;
             if(arg2=get_next_arg(arg1))
              if(arg3=get_next_arg(arg2))
                 {
                   vector = identify_vector(arg1, TRUE);
                   if(vector)
                    {
                     hvector = identify_vector(arg2, TRUE);
                     hhvector = init_dest_vector(arg3, vector->el_cnt);
                     if(hvector && hhvector)
                        cross_product(vector, hvector, hhvector);
                     goto return_it;
                    }
                 }
             display_error("Improper arguments of cross");
             goto return_it;
           case FV_EXTVBS:
             if(arg2=get_next_arg(arg1))
               {
                 vector = identify_vector(arg1, TRUE);
                 evaluate(arg2, &a);
                 if(vector)
                   extrapolate_vbs(vector, &a, result);
                 goto return_it;
               }
             display_error("Improper arguments of extrapolate_vbs");
             result->re = result->im = 0;
             goto return_it;
           case FV_EXTSTB:
             if(arg2=get_next_arg(arg1))
               if(arg3=get_next_arg(arg2))
                 {
                   vector = identify_vector(arg1, TRUE);
                   hvector = identify_vector(arg2, TRUE);
                   evaluate(arg3, &a);
                   if(vector && hvector)
                     extrapolate_stb(vector, hvector, &a, result);
                   goto return_it;
                 }
             display_error("Improper arguments of extrapolate_stb");
             result->re = result->im = 0;
             goto return_it;
           case FV_INTPOLY:
             if(arg2=get_next_arg(arg1))
               if(arg3=get_next_arg(arg2))
                  {
                    vector = identify_vector(arg1, TRUE);
                    hvector = identify_vector(arg2, TRUE);
                    evaluate(arg3, &a);
                    if(vector && hvector)
                      interpolate_poly(vector, hvector, &a, result);
                    goto return_it;
                  }
             display_error("Improper arguments of interpolate_poly");
             result->re = result->im = 0;
             goto return_it;
           case FV_LINREG:
             if(arg2=get_next_arg(arg1))
                  {
                    vector = identify_vector(arg1, TRUE);
                    hvector = identify_vector(arg2, TRUE);
                    if(vector && hvector)
                     {
                      lin_reg(vector, hvector, &a, &b, result);
                      assign_variable((UBYTE *) "rega", &b);
                      assign_variable((UBYTE *) "regb", &a);
                     }
                    goto return_it;
                  }
             display_error("Improper arguments of lin_reg");
             result->re = result->im = 0;
             goto return_it;
           case FV_FOURIER:
             if(arg2=get_next_arg(arg1))
               if(arg3=get_next_arg(arg2))
                  {
                    vector = identify_vector(arg1, TRUE);
                    hvector = init_dest_vector(arg2, vector->el_cnt/2);
                    hhvector = init_dest_vector(arg3, vector->el_cnt/2);
                    if(vector && hvector && hhvector)
                      finite_fourier(vector, hvector, hhvector, result);
                    goto return_it;
                  }
             display_error("Improper arguments of fourier");
             result->re = result->im = 0;
             goto return_it;
           case FV_CFOURIER:
             if(arg2=get_next_arg(arg1))
                  {
                    vector = identify_vector(arg1, TRUE);
                    hvector = init_dest_vector(arg2, vector->el_cnt);
                    if(vector && hvector)
                      complex_finite_fourier(vector, hvector, result);
                    goto return_it;
                  }
             display_error("Improper arguments of complex_fourier");
             result->re = result->im = 0;
             goto return_it;
           case FV_DIM:					/* Get dimension of a vector */
             vector = identify_vector(arg1, TRUE);
             if(vector)
                result->re = (REAL) vector->el_cnt;
             result->im = 0;
             goto return_it;
           case FV_DIMC:				/* Get dimension of (pre)image of a matrix */
           case FV_DIMR:
             matrix = identify_matrix(arg1);
             if(matrix)
                result->re = (REAL) ((f_id == FV_DIMC)?matrix->columns:matrix->rows);
             result->im = 0;
             goto return_it;
         }

        f_id=identify_marg_function(expression);	/* Ist es eine Funktion mit mehreren Argumenten ? */
        switch(f_id)
         {
           case FMARG_INT:
           case FMARG_INTOPEN:
             if(arg2=get_next_arg(arg1))
               if(arg3=get_next_arg(arg2))
                  if(arg4=get_next_arg(arg3))
                    {
                      evaluate(arg3, &a);
                      evaluate(arg4, &b);
                      integrate(arg1, arg2, &a, &b, (f_id == FMARG_INTOPEN), INTTICKS, result);
                      goto return_it;
                    }
             display_error("Improper arguments of int");
             result->re = result->im = 0;
             goto return_it;
           case FMARG_INTEXTR:
           case FMARG_INTOPENEXTR:
             if(arg2=get_next_arg(arg1))
               if(arg3=get_next_arg(arg2))
                  if(arg4=get_next_arg(arg3))
                    {
                      evaluate(arg3, &a);
                      evaluate(arg4, &b);
                      integrate_extrapolate(arg1, arg2, &a, &b, (f_id == FMARG_INTOPENEXTR), INTTICKS, result);
                      goto return_it;
                    }
             display_error("Improper arguments of int");
             result->re = result->im = 0;
             goto return_it;
           case FMARG_INTCAUCHY:
             if(arg2=get_next_arg(arg1))
               if(arg3=get_next_arg(arg2))
                  if(arg4=get_next_arg(arg3))
                    {
                      evaluate(arg3, &a);
                      evaluate(arg4, &b);
                      integrate_cauchy(arg1, arg2, &a, &b, INTTICKS, result);
                      goto return_it;
                    }
             display_error("Improper arguments of int_cauchy");
             result->re = result->im = 0;
             goto return_it;
           case FMARG_MIN:
             if(*arg1 == '{')
              {
               minmaxlist(arg1, result, 1);
               goto return_it;
              }
             if(arg2=get_next_arg(arg1))
               if(arg3=get_next_arg(arg2))
                    {
                      evaluate(arg3, &a);
                      minmaximize(arg1, arg2, &a, result, 1);
                      goto return_it;
                    }
             display_error("Improper arguments of min");
             result->re = result->im = 0;
             goto return_it;
           case FMARG_MAX:
             if(*arg1 == '{')
              {
               minmaxlist(arg1, result, -1);
               goto return_it;
              } 
             if(arg2=get_next_arg(arg1))
               if(arg3=get_next_arg(arg2))
                    {
                      evaluate(arg3, &a);
                      minmaximize(arg1, arg2, &a, result, -1);
                      goto return_it;
                    }
             display_error("Improper arguments of max");
             result->re = result->im = 0;
             goto return_it;
           case FMARG_BINOMIAL:
             if(arg2=get_next_arg(arg1))
                 {
                   evaluate(arg1, &a);
                   evaluate(arg2, &b);
                   binomial(&a, &b, result);
                   goto return_it;
                 }
             display_error("Improper arguments of binomial");
             result->re = result->im = 0;
             goto return_it;
           case FMARG_DF:
             if(arg2=get_next_arg(arg1))
               if(arg3=get_next_arg(arg2))
                    {
                      evaluate(arg3, &a);
                      df(arg1, arg2, &a, DIFFSTEPS, result);
                      goto return_it;
                    }
             display_error("Improper arguments of df");
             result->re = result->im = 0;
             goto return_it;
           case FMARG_DIRECDF:
             if(arg2=get_next_arg(arg1))
               if(arg3=get_next_arg(arg2))
                  if(arg4=get_next_arg(arg3))
                    {
                      evaluate(arg3, &a);
                      evaluate(arg4, &b);
                      diff(arg1, arg2, &a, &b, TRUE, DIFFSTEPS, result);
                      goto return_it;
                    }
             display_error("Improper arguments of direc_df");
             result->re = result->im = 0;
             goto return_it;
         }

        f_id=identify_function(expression);
        if(f_id != F_UNKNOWN)
          evaluate(arg1, &a);		/* Argument auswerten */
        switch(f_id)
         {
           case F_EXP:
             complex_exp(&a, result);
             goto return_it;
           case F_SIN:
             complex_sin(&a, result);
             goto return_it;
           case F_COS:
             complex_cos(&a, result);
             goto return_it;
           case F_TAN:
             complex_tan(&a, result);
             goto return_it;
           case F_COTAN:
             complex_cotan(&a, result);
             goto return_it;
           case F_ASIN:
             complex_asin(&a, result);
             goto return_it;
           case F_ACOS:
             complex_acos(&a, result);
             goto return_it;
           case F_ATAN:
             complex_atan(&a, result);
             goto return_it;
           case F_ACOTAN:
             complex_acotan(&a, result);
             goto return_it;
           case F_SINH:
             complex_sinh(&a, result);
             goto return_it;
           case F_COSH:
             complex_cosh(&a, result);
             goto return_it;
           case F_TANH:
             complex_tanh(&a, result);
             goto return_it;
           case F_COTANH:
             complex_cotanh(&a, result);
             goto return_it;
           case F_ASINH:
             complex_asinh(&a, result);
             goto return_it;
           case F_ACOSH:
             complex_acosh(&a, result);
             goto return_it;
           case F_ATANH:
             complex_atanh(&a, result);
             goto return_it;
           case F_ACOTANH:
             complex_acotanh(&a, result);
             goto return_it;
           case F_ABS:
             result->re = complex_abs(&a);
             result->im = 0.0;
             goto return_it;
           case F_RE:
             result->re = complex_re(&a);
             result->im = 0.0;
             goto return_it;
           case F_IM:
             result->re = complex_im(&a);
             result->im = 0.0;
             goto return_it;
           case F_CC:
             complex_conjugate(&a, result);
             goto return_it;
           case F_ARG:
             result->re = complex_arg(&a);
             result->im = 0.0;
             goto return_it;
           case F_LN:
             complex_ln(&a, result);
             goto return_it;
           case F_SQRT:
             complex_sqrt(&a, result);
             goto return_it;
           case F_BESS_I0:
             bessel_I0(&a, result);
             goto return_it;
           case F_BESS_K0:
             bessel_K0(&a, result);
             goto return_it;
           default:
             function = find_function(expression);			/* Ist es eine benutzerdefinierte Funktion ? */
             if(function)
              {
                evaluate(arg1, &a);
                evaluate_function(function, &a, result);
                goto return_it;
              }
             vector = identify_vector(expression, FALSE);		/* Ist es ein Vektor ? */
             if(vector)
              {
                expression[i] = '(';
                expression[ofs] = ')';
                if(! get_vector_var(expression, temp_str, &ofs))
                    goto return_it;
                ofs--;
                if((ofs < 0) || (ofs >= vector->el_cnt))
                 {
                   display_error("Vector index out of range");
                   goto return_it;
                 }
                result->re = vector->elements[ofs].re;
                result->im = vector->elements[ofs].im;
                goto return_it;
              }
             matrix = identify_matrix(expression);		/* Ist es eine Matrix ? */
             if(matrix)
              {
                expression[i] = '(';
                expression[ofs] = ')';
                if(! get_matrix_var(expression, temp_str, &ofs, &ofs1))
                    goto return_it;
                get_matrix_el(matrix, ofs, ofs1, result);
                goto return_it;
              }
             display_error_text("Unknown function '%s'", expression);
             goto return_it;
         }
       }
   }
  
  /* Variablen */
  charac=conv_to_lower(*expression);
  if(((charac >= 'a') && (charac <= 'z')) || (charac == '%') || (charac == '_'))
   {
     ptr=find_variable(expression);
     if(ptr)
      {
       result->re=ptr->re;
       result->im=ptr->im;
       goto return_it;
      }
     display_error_text("Unknown variable '%s'", expression);
     goto return_it;
   }
  
  /* Wenn es sonst nichts ist, kann es nur eine Zahl sein */
  /* Zahl lesen mit Fehlerueberpruefung */
  result->re=result->im=0.0;
  i=exponent=0;
  p_flag=neg_flag = FALSE;
  if(expression[0] == '-')
     neg_flag=TRUE;
  if((expression[0] == '+') || (expression[0] == '-'))
     i++;
  while(expression[i])
   {
    if(expression[i] == '.')
     {
       if(p_flag)
         display_error_text("Two points in '%s'", expression);
       p_flag=TRUE;
     }
    else
     {
      if(conv_to_lower(expression[i]) == 'e')
        {
          i++;
          sscanf((char *) (expression+i), "%d", &t_exponent);
          exponent += t_exponent;
          if(expression[i] == '+')		/* Fehlerueberpruefung */
             i++;
          else
             if(expression[i] == '-')
                 i++;
          while((expression[i] <= '9') && (expression[i] >= '0'))
               i++;
          if(expression[i])
            display_error("Exponent of '%s': number expected", expression);
          expression[i]=0;
          i--;
        }
      else
       {
         if((expression[i] < '0') || (expression[i] > '9'))
           display_error_text("Expression '%s': number expected", expression);
         result->re *= 10.0;
         result->re += (expression[i] - '0');
         if(p_flag)
           exponent--;
       }
     }
    i++;
   }

  if(neg_flag)
     result->re = -result->re;
  if(exponent < 0)
     for(i=0; i>exponent; i--)
          result->re /= 10.0;
  if(exponent > 0)
     for(i=0; i<exponent; i++)
          result->re *= 10.0;

return_it:
  free(expression);			/* Speicher wieder freigeben */
 }

/* System-Aufruf */

void system_shell(UBYTE *str)
 {
   str = strip_file(str);
   system((char *) str);
 }

/* Eine Funktion definieren */

void process_let(UBYTE *arg, int output_flag)
 {
   UBYTE var_name[200];
   UBYTE f_name[200];
   UBYTE f_def[FDEFLEN];
   int ofs;

   if(DEBUGFLAG > 2)
    printf("# Processing  let '%s'\n", arg);

   while(*arg && (*arg <= ' '))
       arg++;

   ofs=0;
   while(*arg > ' ')
     var_name[ofs++] = *arg++;
   var_name[ofs] = 0;

   while(*arg && (*arg <= ' '))
       arg++;
   if(! *arg)
      goto syntax_error;

   ofs=0;
   while(*arg > ' ')
     f_name[ofs++] = *arg++;
   f_name[ofs] = 0;
   if(! tcompare(f_name, (UBYTE *) "let"))
      goto syntax_error;

   while(*arg && (*arg <= ' '))
       arg++;
   if(! *arg)
      goto syntax_error;

   ofs=0;
   while(*arg && (*arg != '('))
    {
     if(*arg > ' ')
       f_name[ofs++] = *arg;
     arg++;
    }
   f_name[ofs] = 0;
   if(! *arg++)
      goto syntax_error;

   ofs=0;
   while(*arg && (*arg != ')'))
    {
     if(*arg > ' ')
       f_def[ofs++] = *arg;
     arg++;
    }
   f_def[ofs] = 0;
   if(! *arg++)
      goto syntax_error;
   if(! compare(var_name, f_def))
     {
       display_error_text("Argument of function '%s' does not match definition", f_name);
       return;
     }  

   while(*arg && (*arg <= ' '))
       arg++;
   if(*arg++ != '=')
    {
      display_error_text("Missing '=' in definition of function '%s'", f_name);
      return;
    }

   ofs=0;
   while(*arg)
    {
     if(*arg > ' ')
       f_def[ofs++] = *arg;
     arg++;
    }
   f_def[ofs] = 0;
   if(ofs)
    if((f_def[ofs-1] == ';') || (f_def[ofs-1] == '$'))
       f_def[ofs-1] = 0;

   define_function(f_name, var_name, f_def);		/* Funktion definieren */
   if(output_flag)
     printf("    for all %s let %s(%s) = %s\n", var_name, f_name, var_name, f_def);

   return;
 syntax_error:
   display_error_text("Wrong syntax in 'for all %s ...'", var_name);
 }


/* Schleife auswerten */

#define DOOPTION 0xa1
#define SUMOPTION 0xa2
#define PRODOPTION 0xa3

void process_for(UBYTE *arg, COMPLEX *result, int output_flag)
 {
   int cnt_var, cnt_begin, cnt_end;
   UBYTE cnt_name[200];
   UBYTE c, hstring[200], hstring1[STRSIZE];
   register int i, j, k;
   int pos, flag, oflag, zflag;
   int ofs, step;
   COMPLEX temp, temp1;

   if(DEBUGFLAG > 2)
    printf("# Processing  for '%s'\n", arg);

   result->re=result->im=0;

   ofs=0;
   for(i=0; arg[i] && (arg[i] != ':'); i++)
    if(arg[i] > ' ')
      cnt_name[ofs++] = arg[i];
    else
     {
      cnt_name[ofs] = 0;
      if(tcompare(cnt_name, (UBYTE *) "all"))
        {
           process_let(arg+i+1, output_flag);
           return;
        }
     }
   cnt_name[ofs] = 0;
   if(arg[i++] != ':')
     goto error1;
   if(arg[i++] != '=')
     goto error1;
   i = get_number(arg,i, &cnt_begin);
   if(arg[i] == ':')
    {
     i++;
     step = 1;
    }
   else
    {
      while(arg[i] && (arg[i] <= ' '))
            i++;
      if(strncasecmp(arg+i, "step", 4))
           goto error2;
      i = get_number(arg, i+4, &step);
      while(arg[i] && (arg[i] <= ' '))
            ;
      if(strncasecmp(arg+i, "until", 5))
           goto error4;
      i += 5;
    }
   i = get_number(arg,i, &cnt_end);
   if(! strncasecmp(arg+i, "do", 2))
    {
     oflag = DOOPTION;
     arg += 2;
    }
   else
    {
     if(! strncasecmp(arg+i, "sum", 3))
      {
       oflag = SUMOPTION;
       arg += 3;
      }
     else
      {
       if(! strncasecmp(arg+i, "product", 7))
        {
         oflag = PRODOPTION;
         arg += 7;
        }
       else
         goto error3;
      }
    }

   if(oflag == PRODOPTION)
     result->re=1.0;

   if(cnt_begin <= cnt_end)
      step *= 1;
   else
      step *= -1;
   for(cnt_var=cnt_begin; step*(cnt_end-cnt_var)>= 0; cnt_var += step)
    {
     sprintf(hstring, "%d", cnt_var);
     pos = 0;
     zflag = TRUE;
     for(j=i; arg[j]; j++)
      {
        flag=FALSE;
        if(j>0)
         if(((arg[j-1] | 0x20) < 'a') || ((arg[j-1] | 0x20) > 'z'))
           {
             flag=TRUE;
             for(k=0; (k<ofs) && flag; k++)
               if(conv_to_lower(arg[j+k]) != conv_to_lower(cnt_name[k]))
                         flag = FALSE;
             if(flag)
               if(((arg[j+k] | 0x20) >= 'a') && ((arg[j+k] | 0x20) <= 'z'))
                         flag = FALSE;
             if(flag)
              {
                k=0;
                while(hstring1[pos] = hstring[k++])
                        pos++;
                j += (ofs-1);
                zflag = FALSE;
              }
           }
        if(! flag)
         {
          c = arg[j];
          if((c > ' ') || (! zflag))
            hstring1[pos++] = c;
          zflag = (c <= ' ')?TRUE:FALSE;
         }
      }
     hstring1[pos] = 0;
     interpret_line(hstring1, &temp, output_flag);
     switch(oflag)
      {
       case PRODOPTION:
         temp1.re = result->re;
         temp1.im = result->im;
         multiply(&temp1, &temp, result);
         break;
       case SUMOPTION:
         result->re += temp.re;
         result->im += temp.im;
         break;
      }
    }
   return;

error1:
   display_error("After for: No ':=' found");
   return;
error2:
   display_error("After for: ':' or 'step' expected after first number");
   return;
error3:
   display_error("After for: 'do', 'product' or 'sum' expected after last number");
   return;
error4:
   display_error("After for with step: 'until' expected after second number");
   return;
 }

/* Process "if" */

static void depth_warn(UBYTE *s, int depth)
 {
  if(depth < 0)
    display_error_text("Too many ')' in '%s'", s);
  else
    display_error_text("Not enough ')' in '%s'", s);
 }

void process_if(UBYTE *arg, COMPLEX *result, int output_flag)
 {
   UBYTE hstring[STRSIZE];
   int i, j;
   int cflag, neqflag;
   int ofs, lofs, pos, depth;
   COMPLEX lcomp, rcomp, temp;

   if(DEBUGFLAG > 2)
    printf("# Processing  if '%s'\n", arg);

   result->re=result->im=0;

   ofs=0;
   depth = 0;
   for(i=0; arg[i] && (((arg[i] != '=') && (arg[i] != '!')) || depth); i++)
    {
     hstring[ofs++] = arg[i];
     if(arg[i] == '(')
       depth++;
     if(arg[i] == ')')
       depth--;
    }
   if(depth)
     depth_warn(arg, depth);
   hstring[ofs] = 0;

   neqflag = FALSE;		/* "=" sign */
   if(arg[i] == '!')
    {
     neqflag = TRUE;		/* or "!=" sign ? */
     i++;
    }

   if(arg[i++] != '=')
     goto error1;
   lofs = i;

/*
   if(hstring[0] == '(')
    {
     for(; ofs && (hstring[ofs] <= ' '); ofs--)
             ;
     if(hstring[ofs] == ')')
       hstring[ofs] = 0;
     interpret_line(hstring+1, &lcomp, FALSE);
    }
   else
*/
     interpret_line(hstring, &lcomp, FALSE);

   ofs=0;
   depth = 0;
   for(; arg[i] && (strncasecmp(arg+i, "then", 4) || depth); i++)
    {
     hstring[ofs++] = arg[i];
     if(arg[i] == '(')
       depth++;
     if(arg[i] == ')')
       depth--;
    }
   if(depth)
     depth_warn(arg+lofs, depth);
   hstring[ofs] = 0;

/*
   if(hstring[0] == '(')
    {
     for(; ofs && (hstring[ofs] <= ' '); ofs--)
             ;
     if(hstring[ofs] == ')')
       hstring[ofs] = 0;
     interpret_line(hstring+1, &rcomp, FALSE);
    }
   else
*/
     interpret_line(hstring, &rcomp, FALSE);
 
   if(strncasecmp(arg+i, "then", 4))
     goto error2;
   i += 4;
 
   sub(&lcomp, &rcomp, &temp);
   if(complex_abs(&temp) <= ZEROVAL)
      cflag = TRUE;
   else
      cflag = FALSE;

   if(neqflag)			/* Are we testing for NOT equal ? */
      cflag = (! cflag);

   if(cflag)
    {
     if(DEBUGFLAG > 2)
       printf("# if-condition true\n");
     pos = 0;
     for(j=i; hstring[pos++] = arg[j]; j++)
              ;
     interpret_line(hstring, result, output_flag);
    }
   else
     if(DEBUGFLAG > 2)
       printf("# if-condition false\n");
   return;

error1:
   display_error("After if: No '=' found");
   return;
error2:
   display_error("After if: No 'then' found");
   return;
 }

/* Eine Eingabe-Zeile auswerten */

void interpret_line(UBYTE *input, COMPLEX *result, int output_flag)
 {
   int ofs;
   int rows, columns;
   register int i, j, k;
   UBYTE s0, s1;
   UBYTE string[STRSIZE];
   UBYTE hstring[200];
   struct vector *vec;
   struct matrix *mat;
   int arg1, arg2, arg3, arg4;
   UBYTE *argv[MAXARGOUT];
   int argc;
   int sp_flag, cl_flag;
   int temp;
   int bra_level, last_ofs;

   if(DEBUGFLAG)
     printf("# Interpreting line: '%s'\n", input);

   while(*input && (*input <= ' '))
        input++;
   if(! *input)
        return;

   if(*input == '{')
    {
      bra_level=1;
      last_ofs=1;
      for(i=1; input[i]; i++)
       {
         switch(input[i])
          {
            case '{':
              bra_level++;
              break;
            case '}':
              bra_level--;
              if(! bra_level)
               {
                 input[i]=0;
                 interpret_line(input+last_ofs, result, output_flag);
                 last_ofs=i;
               }
              break;
            case ';':
            case '$':
              if(bra_level == 1)
               {
                 input[i]=0;
                 interpret_line(input+last_ofs, result, output_flag);
                 last_ofs=i+1;
               }
              break;
          }
       }
      return;
    }

   sp_flag = FALSE;		/* Kein Leerzeichen bisher */

   i=ofs=0;
   while(input[ofs])
    {
     if(input[ofs] > ' ')
        string[i++]=input[ofs];
     else
      if(! sp_flag)
       {
        string[i] = 0;
        sp_flag=TRUE;		/* Da war das erste Leerzeichen */
	s0 = (string[0] | 0x20);	/* First character of string */
	s1 = (string[1] | 0x20);	/* Second character of string */
	if(s0 == 'r')
	 if(s1 == 'e')
	  {
          if(tcompare(string+2, (UBYTE *) "ad"))	/* Daten fuer Z_4 Modelle einlesen */
           {	/* "read" */
             arg1 = seek_next_arg(input);
             arg2 = seek_next_arg(input+arg1);
             j=strlen(input+arg1+arg2);
             if(input[arg1+arg2+j-1] == ';')
                input[arg1+arg2+j-1] = 0;
             read_z4(input+arg1+arg2, input+arg1);
             return;
           }
          if(tcompare(string+2, (UBYTE *) "adp"))	/* Daten fuer Dispersions-Relationen eines Z_3 Modells einlesen */
           {	/* "readp" */
             arg1 = seek_next_arg(input);
             arg2 = seek_next_arg(input+arg1);
             j=strlen(input+arg1+arg2);
             if(input[arg1+arg2+j-1] == ';')
                input[arg1+arg2+j-1] = 0;
             read_z3p(input+arg1+arg2, input+arg1);
             return;
           }
	  if(tcompare(string+2, (UBYTE *) "adind"))        /* Indizierten Vektor lesen */
           {	/* "readind" */
             arg1 = seek_next_arg(input);
             arg2 = seek_next_arg(input+arg1);
             read_ind(input+arg1, input+arg1+arg2);
             return;
           }
          if(tcompare(string+2, (UBYTE *) "advecs"))       /* Zwei Vektoren */
           {	/* "readvecs" */
             arg1 = seek_next_arg(input);
             arg2 = seek_next_arg(input+arg1);
             arg3 = seek_next_arg(input+arg1+arg2);
             read_vecs(input+arg1, input+arg1+arg2, input+arg1+arg2+arg3);
             return;
           }
	  }
	if(s0 == 's')
	 {
          if(tcompare(string+1, (UBYTE *) "ystem"))	/* System-Aufruf */
           {	/* "system" */
             arg1 = seek_next_arg(input);
             system_shell(input+arg1);
             return;
           }
          if(tcompare(string+1, (UBYTE *) "howvec"))
           {	/* "showvec" */
             i = strlen(input)-1;
             if((input[i] == ';') || (input[i] == '$'))
                 input[i] = 0;
             arg1 = seek_next_arg(input);
             vec = identify_vector(input+arg1, TRUE);
             if(vec)
                show_vector(vec);
             return;
           }
          if(tcompare(string+1, (UBYTE *) "howmat"))
           {	/* "showmat" */
             i = strlen(input)-1;
             if((input[i] == ';') || (input[i] == '$'))
                 input[i] = 0;
             arg1 = seek_next_arg(input);
             mat = identify_matrix(input+arg1);
             if(mat)
                show_matrix(mat);
             return;
           }
	  if(tcompare(string+1, (UBYTE *) "et"))		/* Settings aendern */
           {	/* "set" */
              arg1 = seek_next_arg(input);
              arg2 = seek_next_arg(input+arg1);
              get_number(input, arg1+arg2, &temp);
              set_it(input+arg1, temp);
              result->re=result->im=0;
              return;
           }
	 }
	if(s0 == 'w')
	 if(s1 == 'r')
	  {
          if(tcompare(string+2, (UBYTE *) "ite"))	/* TeX-Tabelle fuer z4 schreiben */
           {	/* "write" */
             arg1 = seek_next_arg(input);
             write_table(input+arg1);
             return;
           }
          if(tcompare(string+2, (UBYTE *) "iteplot"))	/* Gnu-Plotfile fuer Z_4 schreiben */
           {	/* "writeplot" */
             arg1 = seek_next_arg(input);
             write_plot(input+arg1);
             return;
           }
          if(tcompare(string+2, (UBYTE *) "iteiplot"))	/* Gnu-Plotfile fuer Z_3-Impuls-Daten schreiben */
           {	/* "writeiplot" */
             arg1 = seek_next_arg(input);
             write_z3_plot(input+arg1);
             return;
           }
          if(tcompare(string+2, (UBYTE *) "itepplot"))     /* Gnu-Plotfile fuer Impuls-Daten schreiben (P = 0, 2 pi) */
           {	/* "writepplot" */
             arg1 = seek_next_arg(input);
             write_P_plot(input+arg1);
             return;
           }
	  }
	if(s0 == 'o')
	 if(s1 == 'u')
	  {
           if(tcompare(string+2, (UBYTE *) "tplot"))	/* Zwei Vektoren in Gnuplot-File schreiben */
            {	/* "outplot */
              arg1 = seek_next_arg(input);
              arg2 = seek_next_arg(input+arg1);
              arg3 = seek_next_arg(input+arg1+arg2);
              output_plot(input+arg1, input+arg1+arg2, input+arg1+arg2+arg3);
              return;
            }
           if(tcompare(string+2, (UBYTE *) "tput"))
            {	/* "output" */
              arg1 = seek_next_arg(input);
              argv[argc = 0] = input+arg1;
              while(*argv[argc] && (argc < MAXARGOUT))
               {
                arg1 += seek_next_arg(input+arg1);
                argv[++argc] = input+arg1;
               }
              output_Nexpr(argc, (char **) argv);
              return;
            }
	  }
        if(tcompare(string, (UBYTE *) "if"))		/* Process "if" */
         {
            arg1 = seek_next_arg(input);
            process_if(input+arg1, result, output_flag);
            return;
         }
        if(tcompare(string, (UBYTE *) "for"))		/* For auswerten */
         {
            arg1 = seek_next_arg(input);
            process_for(input+arg1, result, output_flag);
            return;
         }
        if(tcompare(string, (UBYTE *) "matrix"))
         {
            if( get_matrix_var(input+ofs, hstring, &rows, &columns) )
               alloc_matrix(hstring, rows, columns);
            if(output_flag)
               printf("   matrix %s(%d,%d)\n", hstring, rows, columns);
            result->re=result->im=0;
            return;
         }
        if(tcompare(string, (UBYTE *) "vector"))
         {
            if( get_vector_var(input+ofs, hstring, &rows) )
               alloc_vector(hstring, rows);
            if(output_flag)
               printf("   vector %s(%d)\n", hstring, rows);
            result->re=result->im=0;
            return;
         }
        if(tcompare(string, (UBYTE *) "in"))
         {
            arg1 = seek_next_arg(input);
            read_file(input+arg1, output_flag);
            if(output_flag)
               printf("   ... end of file\n");
            result->re=result->im=0;
            return;
         }
        if(tcompare(string, (UBYTE *) "help"))		/* Hilfe */
         {
            arg1 = seek_next_arg(input);
            help(input+arg1);
            result->re=result->im=0;
            return;
         }
        if(tcompare(string, (UBYTE *) "apropos"))	/* Apropos */
         {
            arg1 = seek_next_arg(input);
            apropos(input+arg1);
            result->re=result->im=0;
            return;
         }
       }
     ofs++;
    }
   string[i]=0;

   for(i=0; string[i]; i++)
    if(! strncasecmp(string+i, (UBYTE *) "mkid(", 5))
     {
      cl_flag = FALSE;
      k = i;
      for(j=i+5; string[j]; j++)
       {
        if(cl_flag)
          string[k++] = string[j];
        else
         {
          switch(string[j])
           {
             case ',':
               break;		/* Ignore commas */
             case ')':
               cl_flag = TRUE;	/* This is the closing bracket */
               break;
             default:
               string[k++] = string[j];
               break;
           }
         }
       }
      string[k] = 0;
      if(! cl_flag)
        display_error("No closing bracket found after mkid(");
     }

   ofs=i;
   if(ofs > 0)				/* Semikolon am Ende wegschmeissen */
     if((string[ofs-1] == ';') || (string[ofs-1] == '$'))
         string[ofs-1] = 0;

  if(tcompare(string, (UBYTE *) "showfunc"))	/* Funktionen anzeigen */
   {
      show_functions();
      result->re=result->im=0;
      return;
   }

  if(tcompare(string, (UBYTE *) "help"))	/* Hilfe */
    {
      help((UBYTE *) "");
      result->re=result->im=0;
      return;
    }

  if(tcompare(string, (UBYTE *) "set"))		/* Settings anzeigen */
    {
      set_it((UBYTE *) "", 0);
      result->re=result->im=0;
      return;
    }

   if(tcompare(string, (UBYTE *) "bye") || tcompare(string, (UBYTE *) "quit"))
					/* Ende ? */
    {
      output_close();
      free_variables();
      free_matrices();
      free_vectors();
      free_functions();
      exit(0);
    }

   for(i=0; string[i]; i++)
     if(string[i] == ':')
         if(string[i+1] == '=')
             {				/* Variablen-Zuweisung */
               if(! strncasecmp(string+i+2, "for", 3))		/* For auswerten */
                {
                 for(arg1=i+2; strncasecmp(input+arg1, "for", 3); arg1++)
                        ;
                 arg1 += 3;
/* *** Warning: This construction is funny */
                 while(input[arg1] && (input[arg1] <= ' '))	/* Skip spaces */
                      input[arg1++] = 0;
                 process_for(input+arg1, result, FALSE);
                }
               else
                 evaluate(string+i+2, result);
               string[i] = 0;
               for(i=0; string[i]; i++)
                 if(string[i] == '(')		/* Ein Matrix-Element oder Vektor-Komponente */
                   {
                    for(j=i+1; string[j]; j++)
                     if(string[j] == ',')		/* Eine Matrix hat zwei Argumente */
                      {			/* Zuweisung eines Matrix-Elements */
                       if(get_matrix_var(string, hstring, &rows, &columns))
                        {
                         assign_matrix_el(hstring, rows, columns, result);
                         if(output_flag)
                          {
                             printf("    %s := ", string);
                             show_complex(result);
                          }
                        }
                       return;
                      }
                    if(get_vector_var(string, hstring, &rows))	/* Zuweisung einer Vektor-Komponente */
                     {
                      assign_vector_el(hstring, rows, result);
                      if(output_flag)
                       {
                          printf("    %s := ", string);
                          show_complex(result);
                       }
                     }
                    return;
                   }
               assign_variable(string, result);
               if(output_flag)
                {
                  printf("    %s := ", string);
                  show_complex(result);
                }
               return;
             }

   evaluate(string, result);

   if(output_flag)
    {
     printf("    %s = ", string);
     show_complex(result);
    }
 }

/* Anfuehrungszeichen bei einem Dateinamen entfernen */

UBYTE *strip_file(UBYTE *file)
 {
   register int i;

   if(*file == '"')
       file++;

   for(i=0; file[i]; i++)
          ;

   if(i>0)
     if((file[i-1] == ';') || (file[i-1] == '$'))
      {
        i--;
        file[i] = 0;
      }

   if(i>0)
     if(file[i-1] == '"')
        file[i-1] = 0;

   return(file);
 }

/* ****************************************
   Das naechste gueltige Zeichen lesen */

UBYTE readchar(FILE *fp)
 {
   int c;

loop:
   if((c = fgetc(fp)) == EOF)
    {
      display_error("Read past EOF");
      return(0);
    }
   if(c < ' ')
     goto loop;

   if(c == '%')
    {
      while((c != '\n') && (c != EOF))
          c = fgetc(fp);
      goto loop;
    }
   return((UBYTE) c);
 }

/* Eine Eingabedatei lesen */

void read_file(UBYTE *infile, int output_flag)
 {
   FILE *fpi;
   int ofs, bra_level;
   UBYTE instr[STRSIZE];
   COMPLEX result;

   infile = strip_file(infile);

   fpi=fopen((char *) infile, "r");
   if(! fpi)
    {
      display_error_text("Failed to open '%s' for input", infile);
      return;
    }

   ofs = 0;
   bra_level = 0;
   while(1)
    {
     instr[ofs]=readchar(fpi);
     if(! instr[ofs])
         return;
     if(instr[ofs] == '{')
         bra_level++;
     if(instr[ofs] == '}')
         bra_level--;
     if(bra_level < 0)
      {
        display_error("Too many '}'");
        return;
      }
     if((! bra_level) && ofs && ((instr[ofs] == ';') || (instr[ofs] == '$')))
      {
       instr[ofs+1]=0;
       ofs=0;
       if(compare(instr, (UBYTE *) ";end;"))
        {
         fclose(fpi);
         return;
        }
       if((instr[0] | 0x20) == 'o')
        {
         if(compare(instr, (UBYTE *) "off echo;"))
           output_flag = FALSE;
         else
          {
            if(compare(instr, (UBYTE *) "on echo;"))
              output_flag = TRUE;
            else
              interpret_line(instr, &result, output_flag);
          }
        }
       else
         interpret_line(instr, &result, output_flag);
      }
     else
       ofs++;
    }
 }
