/* Standard-Definitions */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <limits.h>
#include <float.h>
/* Definitions for that program */
#include "useful.h"
#include "complex.h"
#include "variables.h"
#include "text.h"
#include "evaluate.h"
#include "functions.h"

/* List of all functions */

struct function *flist = 0;

/* *******************************************
   Verwaltung von Funktionen */

/* Alle Funktionen freigeben */

void free_functions()
 {
   struct function *ptr, *hptr;

   for(ptr=flist; ptr; ptr=hptr)
    {
      hptr=ptr->next;
      free(ptr->name);
      free(ptr->argument);
      free(ptr->definition);
      free(ptr);
    }
   flist=0;
 }

/* Eine Funktion allocatieren */

void define_function(UBYTE *fname, UBYTE *farg, UBYTE *fdef)
 {
   UBYTE *htext;
   UBYTE fdef_temp[FDEFLEN];
   struct function *ptr;
   register int i, j;
   int flag, ofs;

   ofs=0;
   for(i=0; fdef[i]; i++)
    {
      flag=TRUE;
      if(i)
       if(((fdef[i-1] | 0x20) >= 'a') && ((fdef[i-1] | 0x20) <= 'z'))
          flag=FALSE;
      if(flag)
        {
         for(j=0; farg[j] && flag; j++)
           if(conv_to_lower(fdef[i+j]) != conv_to_lower(farg[j]))
                 flag = FALSE;
         if(flag)
           if(((fdef[i+j] | 0x20) >= 'a') && ((fdef[i+j] | 0x20) <= 'z'))
                 flag = FALSE;
        }
      if(! flag)
        fdef_temp[ofs++] = fdef[i];
      else
       {
        fdef_temp[ofs++] = '_';
        for(j=0; fname[j]; j++)
           fdef_temp[ofs++] = fname[j];
        fdef_temp[ofs++] = '_';
        for(j=0; farg[j]; j++)
           fdef_temp[ofs++] = farg[j];
        i += (j-1);
       }
    }
   fdef_temp[ofs] = 0;

   flag=FALSE;
   for(ptr=flist; ptr; ptr=ptr->next)       /* Funktionen ueberschreiben */
    if(compare(fname, ptr->name))
      {
        free(ptr->argument);
        free(ptr->definition);
        flag = TRUE;
        goto eof_loop;
      }

  eof_loop:
   if(! flag)
    {
     htext=malloc(strlen(fname) + 1);              /* Namen speichern */
       if(! htext)
         goto oomem;

     for(i=0; htext[i] = fname[i]; i++)
           ;

     ptr=(struct function *)malloc(sizeof(struct function));
     if(! ptr)
      {
          free(htext);
          goto oomem;
      }
     ptr->name=htext;
    }
   htext=malloc(strlen(farg) + 1);         	     /* Argument speichern */
     if(! htext)
      {
       free(ptr->name);
       free(ptr);
       goto oomem;
      }

   for(i=0; htext[i] = farg[i]; i++)
         ;
   ptr->argument=htext;

   htext=malloc(strlen(fdef_temp) + 1);         	     /* Definition speichern */
     if(! htext)
      {
       free(ptr->argument);
       free(ptr->name);
       free(ptr);
       goto oomem;
      }

   for(i=0; htext[i] = fdef_temp[i]; i++)
         ;
   ptr->definition=htext;
   
   if(! flag)
    {
     ptr->next=flist;
     flist=ptr;
    }
   return;
oomem:
   display_error_text("Out of memory while assigning function '%s'", fname);
 }

/* Eine Funktion suchen */

struct function *find_function(UBYTE *text)
 {
   struct function *ptr;

   for(ptr=flist; ptr; ptr=ptr->next)
    if(compare(ptr->name, text))
       return(ptr);

   return(0);
 }

/* Eine Funktion berechnen */

void evaluate_function(struct function *f, COMPLEX *arg, COMPLEX *result)
 {
   UBYTE var_name[200];
   UBYTE *tmp_ptr;

   sprintf(var_name, "_%s_%s", f->name, f->argument);
   assign_variable(var_name, arg);
   if(*(f->definition) == '{')
    {
       tmp_ptr = (UBYTE *) malloc(strlen(f->definition)+100);
       if(tmp_ptr)
        {
          memcpy(tmp_ptr, f->definition, strlen(f->definition) + 1);
          interpret_line(tmp_ptr, result, FALSE);
          free(tmp_ptr);
        }
       else
          display_error("Out of memory while evaluating block argument of function");
    }
   else
       evaluate(f->definition, result);
 }

/* Alle Funktionen anzeigen */

void show_functions()
 {
   struct function *ptr;
   int fname_len;
   register int i, j;
   int flag;

   printf("The following functions are currently defined by the user:\n");
   for(ptr=flist; ptr; ptr=ptr->next)
    {
      printf("\t%s(%s)\t= ", ptr->name, ptr->argument);
      fname_len=strlen(ptr->name);
      for(i=0; ptr->definition[i]; i++)
       {
         flag=TRUE;
         if(ptr->definition[i] == '_')
           if(! strncmp(ptr->definition+i+1, ptr->name, fname_len))
              if(ptr->definition[i+fname_len+1] == '_')
                 flag=FALSE;
         if(flag)
           putchar(ptr->definition[i]);
         else
           i += (fname_len+1);
       }
      putchar('\n');
    }
 }
