/* *** January 30, 2004: #define LAPACK to replace Householder diagonalization
         routine by LaPack zheev
         (following instruction of Josef Schuele)
   *** October 23, 2004:
    - interfacing with ilaenv modified to improve
      work-size. If possible (e.g. under Linux)
      use zheev itself to determine worksize
      (compile with #define LWORKZHEEV)
    - Post-Householder Gram-Schmidt called only
      if jacobi_flag is set
   *** November 4-5, 2004:
    - try to use standard basis for full diagonalization
      during first pass in order to reduce overhead
      (also delays memory allocation such that less total
       memory is used during full diagonalization)
   *** December 11, 2004:
    - #define READLINE if you want to bind to GNU readline
      library for commandline input
   *** January 20, 2005:
    - bug with fgets() for console input fixed
   *** January 26, 2005:
    - Option '-m' added to compute matrix elements
    - Option '-zo' added to write compressed output
   *** January 28,  2005:
    - Option '-all' suppresses computation of eigenvectors
      when using LAPACK or the Householder algorithm.
      This is useful if you want to do a full diagonalization
      and don't need eigenvectors since it saves CPU time
      and memory in this case
   *** February 2, 2005:
    - Possibility to compute more than eigenvector by
      Lanczos' method (no ghost check, i.e. spurious
      eigenvectors possible with option -b)
    - Eigenvectors produced by Lanczos method can be
      dumped with option -t
   *** February 18, 2005:
    - Lanczos algorithm uses LaPack dsteqr-routine for
      tridiagonalization if compiled with -DLAPACK
   *** March 22, 2005:
    - Parallelization pragma added for eigenvector
      computation in Lanczos' routine
   *** March 25, 2005:
    - Some bugs in recent extensions fixed
*/


/* Das haetten wir gerne noch: */
/* Gleichungssystem loesen (z.B. mit Gauss, ...)
   Inversion von Matrizen (z.B. mit cg, Gauss, ...)
   Matrix-Diagonalisierung mit Eigenwerten in Vektor
   Eigenvektoren, z.B.:
     Itererierter Lanczos-Algorithmus
     Davidson-Algorithmus
     Arnoldi-Algorithmus
     Vektor-Iteration
     ...
   reelle Matrizen
   Berechnung von det's (Gauss, Entwicklungssatz, ...)
   Gram-Schmidt fuer Vektoren
   Funktionen in mehreren Veraenderlichen
   Strings
   Funktionen fitten
   Nullstellen von Fkt.
   Loesen nicht-lin. Glgs-Systeme
   Dlog-Pade

   Statistik: Mittelwerte ...

   Programm-Funktionen (Programm speichern: Labels)
     if, while, ...

   Plot-Funktionen (Kurven, Flaechen, Kreuze, Balkendiagramme ...)
   auf
        Text
        HPLJ II
        X11
        TeX
        Postscript
        PC-Grafik

   Matrix-Rechnung '[matrix] :=' fuer Matrix-Zuweisung,
                         scalar(...), +, -, *, Klammern, numerische Konstanten,
                         @* (Tensor-Produkt), @+ (Direkte Summe),
                         Standard-Einheits-Matrix id(n), adj(matrix) (adjungiert)
   Analog: Vekorrechnung '|vector> :=', ECKIGE Klammern fuer Matrix-Ausdruecke
                               [matrix] (angewandt auf Vektor)
   Man koennte auch |x> fuer den Vektor x nehmen ...
*/

/* Optionen fuer Debugging */

/* #define DEBUG_POW */

/* Matrix-Diagonalisierung */

/* Standard-Definitions */
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <float.h>
#include <math.h>
/* Definitions for that program */
#include "useful.h"
#include "complex.h"
#include "short_alloc.h"
#include "variables.h"
#include "text.h"
#include "matrix.h"
#include "vectors.h"
#include "fullmatrix.h"
#include "extrapolate.h"
#include "evaluate.h"
#include "functions.h"

#include "arguments.h"
#include "help.h"
#include "hamilton.h"
#include "set.h"

#include "date.h"

#ifdef LAPACK
#include "lapack.h"
#endif

#ifdef READLINE
#include <readline/readline.h>
#include <readline/history.h>
#endif

/* Flag, ob Eigenvektoren gespeichert werden sollen */

static int store_eigen = FALSE;

/* Flag if we want to compute eigenvectors */

static int no_eigenvects = FALSE;

/* Flag if input should be piped through gzip */

static int gzip_flag = FALSE;

/* Flag if output should be piped through gzip */
   
static int gzip_out = FALSE;

/* Name of file to store vectors before/after vector iteration in */

static char VIDMPNAME[] = "vi_vecs.dmp";

static char VIDMPNAMEB[128], VIDMPNAMEA[128];	/* - before, after - */

/* Flags whether vectors are to be read from/stored in files
   before or after iteration */

static int read_before = FALSE;
static int write_after = FALSE;

static int skip_first = FALSE;	/* Skip first diagonalization pass ? */

/* Flag whether to compute eigenvectors using the
   Lanczos procedure */

static int lanc_ev_flag = FALSE;

/* Flag whether to compute matrix elements */

static int mel_flag = FALSE;

#define DIAGMEL  123
#define ABOVEMEL 890
#define ALLMEL   456

/* ***********************************************
   Functions to open files that are piped through gzip */

/* open file */

FILE *zip_open(char *fname, char *mode)
 {
  char hstring[256];

  if(*mode == 'r')
   {
    if(! gzip_flag)
      return(fopen(fname, mode));
    sprintf(hstring, "gzip -d -c %s.gz", fname);
   } 
  else
   {
    if(! gzip_out)
      return(fopen(fname, mode));
    sprintf(hstring, "gzip - >%s.gz", fname);
   }
  return(popen(hstring, mode));
 }

/* close file */

int zip_close(FILE *fp, char *mode)
 {
  if(*mode == 'r')
   {
    if(! gzip_flag)
      return(fclose(fp));
   }
  else
   {
    if(! gzip_out)
      return(fclose(fp));
   }
  return(pclose(fp));
 }

/* Matrix diagonalization debug output routine */

#ifdef DEBUGDIAG
void debug_out(char *s)
 {
  printf("%s:\n", s);
  fflush(stdout);
  system("date");
 }
#endif

/* auxiliary allocation routines */

static int dhelp_alloc(struct vector **x, struct vector **y, COMPLEX **value, REAL **diff, int nump)
 {
   *x = (struct vector *)malloc(nump * sizeof(struct vector));
   if(! *x)
     return(FALSE);
   *y = (struct vector *)malloc(nump * sizeof(struct vector));
   if(! *y)
     return(FALSE);
   *value = (COMPLEX *)malign16(nump * sizeof(COMPLEX));
   if(! *value)
     return(FALSE);
   *diff = (REAL *)malloc(nump * sizeof(REAL));
   if(! *diff)
     return(FALSE);

   return(TRUE);
 }
 
static void vhelp_alloc(struct vector *x, int nump, int cols)
 {
#ifdef LAPACK
/* ATTENTION: Be very careful if you use this function:
              * Don't use functions such as move_int_vector()
                which spoil the blocking
              * Use free_Nvectors() to return allocated
                memory */
   alloc_int_vectors(x, nump, cols);
#else
   int j;

   for(j=0; j<nump; j++)
    {
     alloc_int_vector(x+j, cols);
    }
#endif
 }

/* auxiliary functions to store eigenvalues */

static void out_eigen(COMPLEX *value, int num, COMPLEX *stabilize, FILE *outfile)
 {
   register int j;
   COMPLEX temp_complex;
   UBYTE hstring[200];
  
   for(j=0; j<num; j++)
    {
      add(value+j, stabilize, &temp_complex);
      fprintf(outfile, "e%d := ", j);			/* Ergebnisse in Output-File schreiben */
      output_complex(&temp_complex, outfile);
      fprintf(outfile, ";\n");
      sprintf((char *) hstring, "e%d", j);		/* Eigenwerte speichern */
      assign_variable(hstring, &temp_complex);
      if(j)
       {
         sub(value+j, value, &temp_complex);		/* Eigenwertdifferenzen speichern */
         sprintf((char *) hstring, "d%d", j);
         assign_variable(hstring, &temp_complex);
       }
    }
 }

/* Compute elements of a given matrix and write them to output file */

void compute_mel(struct matrix *m, int num, struct vector *x,
   struct vector *y, FILE *outfile)
 {
   register int i, j;
   REAL zero;
   COMPLEX temp;

   fprintf(outfile, "\nmatrix %s(%d,%d);\n\n", m->text, num, num);

   zero = ZEROVAL;

   for(i=0; i<num; i++)
    {
     apply_matrix_nopar(m, x+i, y);
     if(mel_flag != DIAGMEL)
      {
       for(j=0; j<=((mel_flag == ABOVEMEL)?i:(num-1)); j++)
        {
          sprod_nopar(x+j, y, &temp);
	  if(fabs(temp.re) <= zero)
	    temp.re = 0;
	  if(fabs(temp.im) <= zero)
	    temp.im = 0;
          if(complex_abs(&temp) >= zero)
	   {
            fprintf(outfile, "%s(%d,%d) := ", m->text, j+1, i+1);
            output_complex(&temp, outfile);
            fprintf(outfile, ";\n");
	   }
        }
      }
     else
      {
        sprod_nopar(x+i, y, &temp);
	if(fabs(temp.re) <= zero)
	  temp.re = 0;
        if(fabs(temp.im) <= zero)
	  temp.im = 0;
        if(complex_abs(&temp) >= zero)
	 {
          fprintf(outfile, "%s(%d,%d) := ", m->text, i+1, i+1);
          output_complex(&temp, outfile);
          fprintf(outfile, ";\n");
	 }
      }
    }
 }

/* Write vectors to file (if desired) */

static void dump_vectors(int num, int dim, struct vector *v)
 {
   FILE *fp;
   register int j;

   if(! write_after)					/* Write vectors to file ? */
      return;

   fp = fopen(VIDMPNAMEA, "w");
   if(! fp)
    {
     display_error_text("Failed to open file '%s' to store vectors", VIDMPNAMEA);
     write_after = FALSE;
    }
   else
     printf("Writing vectors to file '%s'\n", VIDMPNAMEA);
   fwrite(&num, sizeof(int), 1, fp);
   fwrite(&dim, sizeof(int), 1, fp);
   for(j=0; j<num; j++)
      fwrite(v[j].elements, sizeof(COMPLEX), v[j].el_cnt, fp);
   fclose(fp);
 }

/* ***************************************
   Matrix-Eigenwerte berechnen */

int matrix_eigenvalues(struct matrix *mat, REAL precision, int num, int anum, COMPLEX *stabilize,
                       int jacobi_flag, FILE *outfile)
 {
   struct vector *x, *y;
   struct vector *temp_vec;
   struct vector temp_eig;
   register int i, j;
   int pass;
   COMPLEX *value;
   REAL *diff, gdiff, dprecision;
   int start_with;
   int sort_flag, sb_flag;
   int nfile, temp_int;
   REAL temp_real;
   COMPLEX temp_complex, *temp_complex_ptr;
   UBYTE hstring[200];
   FILE *fp;
   int nump;
   struct full_matrix m, transf;
   struct matrix *mptr;
#ifdef LAPACK
   struct vector tvec;
   int *smap;
   int mnum, mnum1;
   /* variables needed in external functions
    * jobz: 'N' eigenvalues only    'V' eigenvalues & -vectors
    * uplo: 'U' upper triangle of mat used - else 'L'
    * Func = 'ZHETRD' required in ilaenv
    */
   char jobz='V',uplo='U';
   char Func[10];		/* Need at least 7 characters to include the terminating 0 */
   REAL *rwork, *eigval;
   COMPLEX *work;
   integer info, lda, lwork, zn;
   integer wi, wj;
   int transpose_flag;
#endif

   if(no_eigenvects)			/* No eigenvectors ? */
      num = mat->columns;		/* => compute all eigenvalues */

   if(anum <= 0)
     anum = 0;
   nump = num+anum;			/* Trickreicher Weise iterieren wir anum Eigenvektoren mehr als wir brauchen */
   if(nump > mat->columns)		/* Mehr Eigenvektoren als Spalten gibt es nicht */
      nump = mat->columns;

   if(num > mat->columns)		/* Mehr Eigenvektoren als Spalten gibt es nicht */
      num = mat->columns;
      
   if(num == mat->columns)		/* Full diagonalization ? */
     sb_flag = TRUE;			/* -> try to use standard basis */
   else
     sb_flag = FALSE;

   if(read_before)
     sb_flag = FALSE;

   x=y=0;
#ifdef LAPACK
   smap = 0;
   transpose_flag = FALSE;
#endif
   value=0;
   diff=0;
   if(! sb_flag)
    {
     if(! dhelp_alloc(&x, &y, &value, &diff, nump))
       goto error;
     vhelp_alloc(y, nump, mat->columns);
     vhelp_alloc(x, nump, mat->columns);
    }

   if(read_before)					/* Read vectors from file? */
    {
     fp = fopen(VIDMPNAMEB, "r");
     if(! fp)
      {
       display_error_text("Failed to open file '%s' to read vectors for iteration", VIDMPNAMEB);
       read_before = FALSE;
      }
     else
       printf("Reading vectors from file '%s'\n", VIDMPNAMEB);
    }
   if(read_before)
    {
     fread(&nfile, sizeof(int), 1, fp);
     if(nfile < nump)
      {
       display_error("Less than expected number of vectors in file - using all of them");
      }
     if(nfile > nump)
      {
       display_error("More than expected number of vectors in file - using only requested number");
       nfile = nump;
      }
    }
   if(read_before)
    {
     fread(&temp_int, sizeof(int), 1, fp);
     if(temp_int != mat->columns)
      {
       display_error("Dimension in file doesn't match - not using it");
       read_before = FALSE;
       fclose(fp);
      }
    }

   for(j=0; j<nump; j++)
    {
     if(read_before && (j < nfile))
       fread(x[j].elements, sizeof(COMPLEX), x[j].el_cnt, fp);
     else
       if(! sb_flag)
         for(i=0; i<x[j].el_cnt; i += (j+1))
             x[j].elements[i].re = 1.0;
    }
   if(read_before)
      fclose(fp);

   if(! skip_first)			/* Try to skip a diagonalization pass */
    if(! sb_flag)			/* Not using standard basis ? */
      gr_sch(x, nump, 0);		/* Orthonormieren */
   start_with = 0;			/* Wir fangen bei 0 an */
   dprecision = precision;		/* Eine hoehere Genauigkeit */
   for(i=0; i<nump; i++)
        dprecision *= 0.1;
   gdiff = 1e10;
   pass=1;
   while(gdiff > precision)
    {
     if(! sb_flag)
      {
     if(skip_first)
       {
        skip_first = FALSE;
        for(j=0; j<nump; j++)		/* Copy x to y */
          copy_vector(x+j, y+j);
        goto skip_dest;			/* Try to skip a diagonalization pass */
       }
#ifdef DEBUGDIAG
     debug_out("Started applying matrices before diagonalization");
#endif
#pragma omp parallel for private(i,j)
     for(j=start_with; j<nump; j++)
       for(i=0; i<5; i++)
        {
          apply_matrix_nopar(mat, x+j, y+j);
          sub_vector_nopar(stabilize, x+j, y+j);
          apply_matrix_nopar(mat, y+j, x+j);
          sub_vector_nopar(stabilize, y+j, x+j);
          normalize_vector_nopar(x+j);
        }
#ifdef DEBUGDIAG
     debug_out("Finished applying matrices before diagonalization");
     debug_out("Started estimating eigenvalues before diagonalization");
#endif
#pragma omp parallel for private(j)
     for(j=start_with; j<nump; j++)
      {
       apply_matrix_nopar(mat, x+j, y+j);
       sub_vector_nopar(stabilize, x+j, y+j);
       diff[j] = estimate_eigenvalue(x+j, y+j, value+j);
       apply_matrix_nopar(mat, y+j, x+j);
       sub_vector_nopar(stabilize, y+j, x+j);
      }
#ifdef DEBUGDIAG
     debug_out("Finished estimating eigenvalues before diagonalization");
#endif
     sort_flag=TRUE;			/* Bubble-Sort */
#ifdef LAPACK
     smap = (int *)malloc(nump * sizeof(int));
     if(! smap)
        goto error;
     for(i=0; i<nump; i++)
        smap[i] = i;
#endif
     while(sort_flag)
      {
        sort_flag = FALSE;
        for(j=start_with; j<nump-1; j++)
         if(complex_abs(value+j+1) > complex_abs(value+j))
           {
              temp_real=diff[j];	/* Elemente von diff vertauschen */
              diff[j] = diff[j+1];
              diff[j+1] = temp_real;
              temp_real=value[j].re;	/* Realteil von value vertauschen */
              value[j].re = value[j+1].re;
              value[j+1].re = temp_real;
              temp_real=value[j].im;	/* Imaginaerteil von value vertauschen */
              value[j].im = value[j+1].im;
              value[j+1].im = temp_real;
#ifdef LAPACK
              /* If things are stored at a piece, we cannot
                 swap pointers !!!!!!!!! */
              temp_int = smap[j];
              smap[j] = smap[j+1];
              smap[j+1] = temp_int;
#else
              temp_int = x[j].el_cnt;	/* Vektoren vertauschen */
              x[j].el_cnt = x[j+1].el_cnt;
              x[j+1].el_cnt = temp_int;
              temp_complex_ptr = x[j].elements;
              x[j].elements = x[j+1].elements;
              x[j+1].elements = temp_complex_ptr;
#endif
              sort_flag=TRUE;		/* Flag fuer weiteren Durchgang setzen */
           }
      }					/* Ende Bubble-Sort */
#ifdef LAPACK
     alloc_int_vector(&tvec, mat->columns);
     for(i=0; i<nump; i++)
      if((smap[i] >= 0) && (smap[i] != i))
       {
        copy_vector(x+i, &tvec);
        mnum = i;
        while(smap[mnum] != i)
         {
          mnum1 = smap[mnum];
          copy_vector(x+mnum1, x+mnum);
          smap[mnum] = -1;
          mnum = mnum1;
         }
        copy_vector(&tvec, x+mnum);
	smap[mnum] = -1;
       }
     free_vector(&tvec);
     free(smap);
#endif

     for(j=0; j<nump; j++)
      {
       add(value+j, stabilize, &temp_complex);
       printf("%d. Estimate for Eigenvalue %d: (%17.17g) ", pass, j+1, diff[j]);
       show_complex(&temp_complex);
      }
     if(diff[start_with] <= dprecision)	/* Einen weiteren Vektor nicht veraendern */
        {
          normalize_vector(x + start_with);	/* Noch einmal normieren */
          dprecision *= 10.0;			/* Naechstes Mal muss es nicht so genau werden */
          start_with++;
        }
#ifdef DEBUGDIAG
     debug_out("Gram-Schmidt called");
#endif
     gr_sch(x, nump, start_with);			/* Orthonormieren */
#ifdef DEBUGDIAG
     debug_out("Gram-Schmidt finished");
#endif
      }		/* end of: if(! sb_flag) */

     if(start_with < nump)
      {				/* Beginning of "if(start_with < nump)" */
        if(! sb_flag)
	 {
#pragma omp parallel for private(j)
          for(j=start_with; j<nump; j++)		/* Prepare for explicit diagonalization */
           {
            apply_matrix_nopar(mat, x+j, y+j);		/* Compute images of vectors */
            sub_vector_nopar(stabilize, x+j, y+j);
           }
#ifdef DEBUGDIAG
          debug_out("Started computing matrix");
#endif
         }
#ifdef LAPACK
     alloc_full_matrix(&m, nump-start_with, nump-start_with, TRUE);	/* Allocate m in one block */
#else
     alloc_full_matrix(&m, nump-start_with, nump-start_with, FALSE);	/* Allow splitting rows of m */
#endif
     transf.rows = transf.columns = 0;
     if(! sb_flag)
      {
#pragma omp parallel for private(i,j)
     for(i=0; i<m.rows; i++)				/* Compute matrix */
      for(j=0; j<m.columns; j++)
        sprod_nopar(x+i+start_with, y+j+start_with, m.entries[i]+j);
#ifdef DEBUGDIAG
     debug_out("Finished computing matrix");
#endif
      }
     else
      {
        copy_matrix_to_full(mat, &m);
	for(i=0; i<m.rows; i++)
	  sub(m.entries[i]+i, stabilize, m.entries[i]+i);
      }		/* end of: if(! sb_flag) */
#ifdef DEBUG_POW
     printf("Current matrix = ");
     show_full_matrix(&m);
#endif
     if(jacobi_flag)
      {
       jacobi_diagonalize(&m, &transf, dprecision/((double) 100));
							/* diagonalize it using Jacobi method */
#ifdef DEBUG_POW
       printf("After Jacobi method =\n");
       show_full_matrix(&m);
       printf("Transformation matrix =\n");
       show_full_matrix(&transf);
#endif
      }
     else
      {
#ifdef LAPACK
   /* According to documentation, ilaenv returns with the
    * specified input the optimal NB value needed to get
    * the optimal lwork.
    * However, this doesn't really seem to work and
    * sometimes it's better to use zheev - define LWORKZHEEV
    * to switch definitions.
    */
#ifndef LWORKZHEEV
     sprintf(Func,"%s","ZHETRD");
     wi = 1;
     wj = m.rows;
#ifdef DEBUGDIAG
     debug_out("ilaenv_ called");
#endif
     lwork = (int) ilaenv_(&wi, Func, &uplo, &wj, &wj, &wj, &wj);
#ifdef DEBUGDIAG
     sprintf((char *) hstring, "ilaenv_ finished (optimal lwork=%d)", lwork);
     debug_out((char *) hstring);
#endif
     if(lwork < 32)
       lwork = 32;
     lwork = m.rows*(lwork+1);
#endif

     rwork = (REAL *)malloc(sizeof(REAL)*(3*m.rows));
     if(! rwork)
        goto error;
     eigval = (REAL *)malloc(sizeof(REAL)*m.rows);
     if(! eigval)
        goto error;

     if(no_eigenvects)
        jobz = 'N';
#ifdef LWORKZHEEV
     work=(COMPLEX *)malign16(sizeof(COMPLEX));
     if(! work)
        goto error;

     lwork = -1;
     lda = zn = m.rows;
     info = 0;
#ifdef DEBUGDIAG
     debug_out("zheev_ called to query optimal lwork");
#endif
     /* Query optimal work size */
     zheev_(&jobz, &uplo, &zn, m.entries[0], &lda, eigval, work, &lwork, rwork, &info);

     lwork = (int)work[0].re;
#ifdef DEBUGDIAG
     sprintf((char *) hstring, "zheev_ finished (optimal lwork=%d)", lwork);
     debug_out((char *) hstring);
#endif

     free(work);
#endif

     work=(COMPLEX *)malign16(sizeof(COMPLEX)*lwork);
     if(! work)
        goto error;

     lda = zn = m.rows;
     info = 0;
#ifdef DEBUGDIAG
     debug_out("zheev_ called");
#endif
     zheev_(&jobz, &uplo, &zn, m.entries[0], &lda, eigval, work, &lwork, rwork, &info);
     if(info)
      {
        sprintf((char *) hstring, "In zheev: %d",info);
        display_error(hstring);
      }
     else
      {
       if(lwork < (int)work[0].re)
        {
         sprintf((char *) hstring, "Not optimal work-size used ?? %d vs. %d", lwork, (int)work[0].re);
         display_error(hstring);
        }
      }
#ifdef DEBUGDIAG
     debug_out("zheev_ finished");
#endif
#ifdef DEBUGLAPACK
printf("\nEigenvalues\n");
for(i=0;i<m.rows;i++)
   printf("%.10f\n", eigval[i] + stabilize->re);
printf("\n");
#endif
     if(no_eigenvects)
      {
       free_full_matrix(&m);
       if(sb_flag)
        {
         if(! dhelp_alloc(&x, &y, &value, &diff, nump))
            goto error;
        }
       for(j=0; j<num; j++)
        {
         value[j].re = eigval[j];
         value[j].im = 0;
        }
       out_eigen(value, num, stabilize, outfile);
      }
     free(work);
     free(eigval);
     free(rwork);
     transpose_flag = TRUE;
     if(no_eigenvects)
        goto free_it;
#else
#ifdef DEBUGDIAG
     debug_out("Householder called");
#endif
     if(no_eigenvects)
      {
       alloc_int_vector(&temp_eig, mat->columns);
       householder_diagonalize(&m, &temp_eig);		/* diagonalize it using Householder method */
      }
     else
       householder_diagonalize(&m, 0);			/* diagonalize it using Householder method */
#ifdef DEBUGDIAG
     debug_out("Householder finished");
#endif
     if(no_eigenvects)
      {
       free_full_matrix(&m);
       out_eigen(temp_eig.elements, num, stabilize, outfile);
       free_vector(&temp_eig);
       goto free_it;
      }
       transf.rows = m.rows;
       transf.columns = m.columns;
       transf.entries = m.entries;
#endif			/* end of  #ifdef LAPACK */
      }
     if(sb_flag)
      {
       if(! dhelp_alloc(&x, &y, &value, &diff, nump))
          goto error;
       vhelp_alloc(y, nump, mat->columns);

#ifdef DEBUGDIAG
     debug_out("Copying eigenvectors started");
#endif

#pragma omp parallel for private(i,j,temp_vec)
       for(i=0; i<m.columns; i++)			/* Now apply this transformation */
        {
         temp_vec = y + i + start_with;
         for(j=0; j < temp_vec->el_cnt; j++)
          {
#ifdef LAPACK
             complex_conjugate(m.entries[i]+j, temp_vec->elements+j);
#else
             temp_vec->elements[j].re = transf.entries[j][i].re;
             temp_vec->elements[j].im = transf.entries[j][i].im;
#endif
          }
        }
#ifdef DEBUGDIAG
     debug_out("Copying eigenvectors finished");
#endif
      }
     else
      {
#ifdef DEBUGDIAG
     debug_out("Applying transformation started");
#endif
/* Be careful - in FORTRAN everything is transposed -
   we do not know how to call zgemm_ properly ... */
#ifdef LAPACK
     if(transpose_flag)					/* ... so, do this by hand */
#pragma omp parallel for private(i,j,temp_vec,temp_complex)
       for(i=0; i<m.columns; i++)			/* Now apply this transformation */
        {
         temp_vec = y + i + start_with;
         for(j=0; j < temp_vec->el_cnt; j++)
          {
             temp_vec->elements[j].re = temp_vec->elements[j].im = 0;
          }
         for(j=0; j<m.rows; j++)
          {
           complex_conjugate(m.entries[i]+j, &temp_complex);
           add_vector_nopar(&temp_complex, x+start_with+j, temp_vec);
          }
        }
     else
#endif
#pragma omp parallel for private(i,j,temp_vec)
       for(i=0; i<transf.columns; i++)			/* Now apply this transformation */
        {
         temp_vec = y + i + start_with;
         for(j=0; j < temp_vec->el_cnt; j++)
          {
             temp_vec->elements[j].re = temp_vec->elements[j].im = 0;
          }
         for(j=0; j<transf.rows; j++)
           add_vector_nopar(transf.entries[j]+i, x+start_with+j, temp_vec);
        }
#ifdef DEBUGDIAG
     debug_out("Applying transformation finished");
#endif
      }			/* End of if(sb_flag) */
#ifdef DEBUG_POW
     for(j=start_with; j<nump; j++)
      {
        apply_matrix(mat, y+j, x+j);			/* Compute images of vectors */
        sub_vector(stabilize, y+j, x+j);
      }

     for(i=0; i<m.rows; i++)				/* Compute matrix */
      for(j=0; j<m.columns; j++)
        sprod(y+i+start_with, x+j+start_with, m.entries[i]+j);
     printf("New matrix =\n");
     show_full_matrix(&m);
#endif

     free_full_matrix(&m);				/* Free the matrices */
     if(jacobi_flag)
      {
       free_full_matrix(&transf);
#ifdef DEBUGDIAG
       debug_out("Post-Jacobi Gram-Schmidt called");
#endif
       gr_sch(y, nump, start_with);			/* Noch einmal Orthonormieren */
#ifdef DEBUGDIAG
       debug_out("Post-Jacobi Gram-Schmidt finished");
#endif
      }

     if(sb_flag)					/* allocate the final set of vectors */
      {
       vhelp_alloc(x, nump, mat->columns);
       sb_flag = FALSE;
      }

skip_dest:			/* Go here if skipping diagonalization pass */
#ifdef DEBUGDIAG
     debug_out("Started applying matrices");
#endif
#pragma omp parallel for private(j)
     for(j=start_with; j<nump; j++)
      {
       apply_matrix_nopar(mat, y+j, x+j);
       sub_vector_nopar(stabilize, y+j, x+j);
       diff[j] = estimate_eigenvalue(y+j, x+j, value+j);
      }
#ifdef DEBUGDIAG
     debug_out("Finished applying matrices");
#endif
#ifdef DEBUG_POW
     for(j=0; j<nump; j++)
      {
       add(value+j, stabilize, &temp_complex);
       printf("%d. Estimate for Eigenvalue %d: (%17.17g) ", pass, j+1, diff[j]);
       show_complex(&temp_complex);
      }
#endif

      }				/* End of "if(start_with < nump)" */

     /* We may be able to save a diagonalization pass if we test
        accuracy only after the (optional) explicit diagonalization step */
     gdiff = 0;
     for(j=0; j<num; j++)
       if(diff[j] > gdiff)
           gdiff = diff[j];

     pass++;
    }

   for(j=0; j<nump; j++)
    {
     add(value+j, stabilize, &temp_complex);
     printf("Final Estimate for Eigenvalue %d: (%17.17g) ", j+1, diff[j]);
     show_complex(&temp_complex);
    }

   out_eigen(value, num, stabilize, outfile);		/* Store eigenvalues */

   if(mat->text[0] == '*')				/* Internal matrix application ? */
       compute_correl(x, outfile); 

#pragma omp parallel for private(j)
   for(j=0; j<nump; j++)
     normalize_vector(x+j);				/* alles auf 1 normieren */

   if(mel_flag)
    {
 
      for(mptr = matrix_list; mptr->next; mptr=mptr->next)
       if((mat->columns == mptr->columns) && (mat->rows == mptr->rows))
        if(strncmp(mptr->text, "_house", 6))
	 compute_mel(mptr, num, x, y, outfile);
    }

   dump_vectors(nump, mat->columns, x);			/* Write vectors to file (?) */

   for(j=0; j<nump; j++)
    {
     if((store_eigen) && (j < num))
      {
        sprintf((char *) hstring, "ev%d", j);
        alloc_vector(hstring, x[j].el_cnt);
        temp_vec = identify_vector(hstring, TRUE);
        copy_vector(x+j, temp_vec);
      }
#ifndef LAPACK
     free_vector(x+j);
     free_vector(y+j);
#endif
    }
#ifdef LAPACK
   free_Nvectors(x, nump);
   free_Nvectors(y, nump);
#endif
free_it:
   if(x)
     free(x);
   if(y);
     free(y);
   if(value);
     free(value);
   if(diff)
     free(diff);
   return(num);
error:
   display_error("Not enough space for eigenvalues (vectors)");
   goto free_it;
 }

/* ***************************************
   Eigenwerte einer nicht-hermiteschen Matrix berechnen */
/* We proceed as follows:
   + We use vector iteration
   + We always iterate two vectors at once
   + The Matrix in this 2x2 subspace is diagonalized explicitly
     (in order to ensure good separation of eigenvalues that
      are relatively close)
   + Then the same procedure is applied to the adjoint of the
     matrix
   + Now higher eigenvectors/values can be determined using
     the previous eigenvectors and adjoint eigenvectors to
     project out the previous eigenvalues
*/
/* NOTE:
   The spectral shift should be chosen complex in order to
   avoid eigenvalues of same absolute value (in particular
   complex conjugate ones)
*/
/* (The code below is not very well written for historic
    reasons - sorry for that) */

#define ANAME "_adjoint_m"
#define EXPLICIT

int non_herm_matrix_eigenvalues(struct matrix *mat, REAL precision, int num, COMPLEX *stabilize, FILE *outfile)
 {
   struct vector *x, *y, *z;
   struct vector *temp_vec;
   struct vector *tv;
   register int i, j;
   int pass;
   COMPLEX *value;
   REAL *diff;
   int sort_flag;
   int cur_pos;
   int temp_int;
   int pos1, pos2;
   REAL temp_real;
   REAL ra;
   COMPLEX ca, cb, cc, cd;
   COMPLEX lam1, lam2, ta, tb, tc, td, te;
   COMPLEX h11, h12, h21, h22;
   COMPLEX s12, s21;
   COMPLEX temp_complex, *temp_complex_ptr;
   UBYTE hstring[200];
   int nump;
   struct matrix *amat, *matr;
   int apass;

   alloc_matrix((UBYTE *) ANAME, mat->columns, mat->rows);	/* Adjungierte Matrix bestimmern */
   amat = identify_matrix((UBYTE *) ANAME);
   adjoint_matrix(mat, amat);

   nump = num+1;                        /* Trickreicher Weise iterieren wir einen Eigenvektor mehr als wir brauchen */
   if(nump > mat->columns)              /* Mehr Eigenvektoren als Spalten gibt es nicht */
      nump = mat->columns;

   if(num > mat->columns)               /* Mehr Eigenvektoren als Spalten gibt es nicht */
      num = mat->columns;

   z=x=y=0;
   value=0;
   diff=0;
   x=(struct vector *)malloc(nump * sizeof(struct vector));
   if(! x)
     goto error;
   z=(struct vector *)malloc(nump * sizeof(struct vector));
   if(! z)
     goto error;
   y=(struct vector *)malloc(nump * sizeof(struct vector));
   if(! y)
     goto error;
   value=(COMPLEX *)malign16(nump * sizeof(COMPLEX));
   if(! value)
     goto error;
   diff=(REAL *)malloc(nump * sizeof(REAL));
   if(! diff)
     goto error;

   for(j=0; j<nump; j++)
    {
     alloc_int_vector(x+j, mat->columns);
     alloc_int_vector(z+j, mat->columns);
     alloc_int_vector(y+j, mat->columns);
     for(i=0; i<x[j].el_cnt; i += (j+1))
         x[j].elements[i].re = 1.0;
     normalize_vector(x+j);		/* Normieren */
     for(i=0; i<z[j].el_cnt; i += (j+1))
         z[j].elements[i].re = 1.0;
     normalize_vector(z+j);             /* Normieren */
    }

   for(cur_pos=0; cur_pos < num; cur_pos++)
    for(apass = 0; apass<2; apass++)
    { 
     if(! apass)			/* Abwechselnd Matrix */
        matr = mat;
     else				/* Und adjungierte iterieren */
        matr = amat;
   diff[cur_pos] = 1e10;
   pass=1;
   while(diff[cur_pos] > precision*precision)
    {
#ifdef EXPLICIT
     for(j=cur_pos; (j<nump) && (j < cur_pos+2); j++)
#else
     j = cur_pos;
#endif
      {
       apply_matrix(matr, x+j, y+j);
       sub_vector(stabilize, x+j, y+j);
       for(i=0; i<cur_pos; i++)         /* Bereits gefundene Eigenvektoren herausprojezieren */
        {
          sprod(z+i, x+j, &ta);
          sprod(z+i, x+i, &tb);
          multiply(value+i, &ta, &tc);
          divide(&tc, &tb, &td);
          sub_vector(&td, x+i, y+j);
        }
       apply_matrix(matr, y+j, x+j);
       sub_vector(stabilize, y+j, x+j);
       for(i=0; i<cur_pos; i++)         /* Bereits gefundene Eigenvektoren herausprojezieren */
        {
          sprod(z+i, y+j, &ta);
          sprod(z+i, x+i, &tb);
          multiply(value+i, &ta, &tc);
          divide(&tc, &tb, &td);
          sub_vector(&td, x+i, x+j);
        }
      }
#ifdef EXPLICIT
     for(j=cur_pos; (j < nump) &&(j < cur_pos+2); j++)
#else
     j = cur_pos;
#endif
       normalize_vector(x+j);			/* Normieren */

#ifdef EXPLICIT
     for(j=cur_pos; (j < nump) &&(j < cur_pos+2); j++)
#else
     j = cur_pos;
#endif
      {
       apply_matrix(matr, x+j, y+j);    /* Bilder der Vektoren berechnen */
       sub_vector(stabilize, x+j, y+j);
       for(i=0; i<cur_pos; i++)         /* Bereits gefundene Eigenvektoren herausprojezieren */
        {
          sprod(z+i, x+j, &ta);
          sprod(z+i, x+i, &tb);
          multiply(value+i, &ta, &tc);
          divide(&tc, &tb, &td);
          sub_vector(&td, x+i, y+j);
        }
       diff[j] = estimate_eigenvalue(x+j, y+j, value+j);
      }

#ifdef EXPLICIT
     pos1 = cur_pos;
     pos2 = cur_pos+1;
     if(pos2 < nump)
      {

#ifdef DEBUG_POW
        printf("Explicit diagonalization applied at positions %d, %d\n", pos1+1, pos2+1);
#endif
/* Let the matrix M in the subspace given by |1> = x(pos1)
   and |2> = x(pos2) have the form
     / a  b \
    |        |
     \ c  d /    .
   Then the matrix elements of M are related to scalar products by
     <1|M|1> = a <1|1> + c <1|2>
     <2|M|1> = a <2|1> + c <2|2>
     <1|M|2> = b <1|1> + d <1|2>
     <2|M|2> = b <2|1> + d <2|2>
   This system is evaluated and solved below */
        sprod(x+pos1, y+pos1, &h11);			/* Matrix-Elemente berechnen */
        sprod(x+pos1, y+pos2, &h12);
        sprod(x+pos2, y+pos1, &h21);
        sprod(x+pos2, y+pos2, &h22);
        sprod(x+pos1, x+pos2, &s12);
        sprod(x+pos2, x+pos1, &s21);
        multiply(&s12, &s21, &ta);
        tb.re = 1.0;
        tb.im = 0.0;
        sub(&tb, &ta, &te);
#ifdef DEBUG_POW
        printf("1 - <1|2><2|1> = ");
        show_complex(&te);
#endif
        multiply(&s12, &h21, &ta);
        sub(&h11, &ta, &tb);
        divide(&tb, &te, &ca);
        multiply(&s12, &h22, &ta);
        sub(&h12, &ta, &tb);
        divide(&tb, &te, &cb);
        multiply(&s21, &h11, &ta);
        sub(&h21, &ta, &tb);
        divide(&tb, &te, &cc);
        multiply(&s21, &h12, &ta);
        sub(&h22, &ta, &tb);
        divide(&tb, &te, &cd);
#ifdef DEBUG_POW
        printf("a = ");
        show_complex(&ca);
        printf("b = ");
        show_complex(&cb);
        printf("c = ");
        show_complex(&cc);
        printf("d = ");
        show_complex(&cd);
#endif
/* Let M be as before. Now we compute the eigenvalues of M via:
    lam_1,2 = (a+d)/2 +/- sqrt(((a-d)/2)^2 + b c)
   where we take the negative root for lam_1 and the
   positive for lam_2 */ 
        add(&ca, &cd, &ta);                           /* Eigenwerte berechnen */
        ta.re /= 2.0;
        ta.im /= 2.0;
        multiply(&ta, &ta, &tb);
        multiply(&ca, &cd, &tc);
        multiply(&cb, &cc, &te);
        sub(&te, &tc, &td);
        add(&td, &tb, &tc);
        complex_sqrt(&tc, &tb);
        sub(&ta, &tb, &lam1);
        add(&ta, &tb, &lam2);
#ifdef DEBUG_POW
        printf("lam1 = ");
        show_complex(&lam1);
        printf("lam2 = ");
        show_complex(&lam2);
#endif
/* Now comes the change of basis. We replace
    |1>   ->  |1> + 1/b (lam_1 - a) |2>
   (if b != 0, unchanged otherwise), and
    |2>   ->  |2> + 1/c (lam_2 - d) |1>
   (if c != 0, unchanged otherwise) */
        if(complex_abs_sq(&cc))                       /* Fall cc != 0 */
         {
          sub(&lam2, &cd, &ta);
          divide(&ta, &cc, &h11);
         }
        else
         {
          h11.re = h11.im =0;
         }
        if(complex_abs_sq(&cb))                       /* Fall cb != 0 */
         {
          sub(&lam1, &ca, &ta);
          divide(&ta, &cb, &h12);
         }
        else
         {
          h12.re = h12.im =0;
         }
#ifdef DEBUG_POW
        printf("(lam1-a)/b = ");
        show_complex(&h12);
        printf("(lam2-d)/c = ");
        show_complex(&h11);
#endif
        copy_vector(x+pos1, y+pos1);
        copy_vector(x+pos2, y+pos2);
        add_vector(&h11, y+pos1, x+pos2);
        add_vector(&h12, y+pos2, x+pos1);
        normalize_vector(x+pos1);
        apply_matrix(matr, x+pos1, y+pos1);
        sub_vector(stabilize, x+pos1, y+pos1);
        for(i=0; i<cur_pos; i++)         /* Bereits gefundene Eigenvektoren herausprojezieren */
         {
           sprod(z+i, x+pos1, &ta);
           sprod(z+i, x+i, &tb);
           multiply(value+i, &ta, &tc);
           divide(&tc, &tb, &td);
           sub_vector(&td, x+i, y+pos1);
         }
        diff[pos1] = estimate_eigenvalue(x+pos1, y+pos1, value+pos1);
        normalize_vector(x+pos2);
        apply_matrix(matr, x+pos2, y+pos2);
        sub_vector(stabilize, x+pos2, y+pos2);
        for(i=0; i<cur_pos; i++)         /* Bereits gefundene Eigenvektoren herausprojezieren */
         {
           sprod(z+i, x+pos2, &ta);
           sprod(z+i, x+i, &tb);
           multiply(value+i, &ta, &tc);
           divide(&tc, &tb, &td);
           sub_vector(&td, x+i, y+pos2);
         }
        diff[pos2] = estimate_eigenvalue(x+pos2, y+pos2, value+pos2);
#ifdef DEBUG_POW
        sprod(x+pos1, y+pos1, &h11);                    /* Matrix-Elemente berechnen */
        sprod(x+pos1, y+pos2, &h12);
        sprod(x+pos2, y+pos1, &h21);
        sprod(x+pos2, y+pos2, &h22);
        sprod(x+pos1, x+pos2, &s12);
        sprod(x+pos2, x+pos1, &s21);
        multiply(&s12, &s21, &ta);
        tb.re = 1.0;
        tb.im = 0.0;
        sub(&tb, &ta, &te);
        multiply(&s12, &h21, &ta);
        sub(&h11, &ta, &tb);
        divide(&tb, &te, &ca);
        multiply(&s12, &h22, &ta);
        sub(&h12, &ta, &tb);
        divide(&tb, &te, &cb);
        multiply(&s21, &h11, &ta);
        sub(&h21, &ta, &tb);
        divide(&tb, &te, &cc);
        multiply(&s21, &h12, &ta);
        sub(&h22, &ta, &tb);
        divide(&tb, &te, &cd);
        printf("a = ");
        show_complex(&ca);
        printf("b = ");
        show_complex(&cb);
        printf("c = ");
        show_complex(&cc);
        printf("d = ");
        show_complex(&cd);
#endif
      }

#ifdef DEBUG_POW
     for(pos1 = 0; pos1 < nump-1; pos1++)
      for(pos2 = pos1+1; pos2 < nump; pos2++)
      {
        printf("Matrix elements at positions %d, %d\n", pos1+1, pos2+1);
        sprod(x+pos1, y+pos1, &h11);                    /* Matrix-Elemente berechnen */
        sprod(x+pos1, y+pos2, &h12);
        sprod(x+pos2, y+pos1, &h21);
        sprod(x+pos2, y+pos2, &h22);
        sprod(x+pos1, x+pos2, &s12);
        sprod(x+pos2, x+pos1, &s21);
        multiply(&s12, &s21, &ta);
        tb.re = 1.0;
        tb.im = 0.0;
        sub(&tb, &ta, &te);
        printf("1 - <1|2><2|1> = ");
        show_complex(&te);
        multiply(&s12, &h21, &ta);
        sub(&h11, &ta, &tb);
        divide(&tb, &te, &ca);
        multiply(&s12, &h22, &ta);
        sub(&h12, &ta, &tb);
        divide(&tb, &te, &cb);
        multiply(&s21, &h11, &ta);
        sub(&h21, &ta, &tb);
        divide(&tb, &te, &cc);
        multiply(&s21, &h12, &ta);
        sub(&h22, &ta, &tb);
        divide(&tb, &te, &cd);
        printf("a = ");
        show_complex(&ca);
        printf("b = ");
        show_complex(&cb);
        printf("c = ");
        show_complex(&cc);
        printf("d = ");
        show_complex(&cd);
      }
#endif
     j = cur_pos;
      if(j < nump-1)
        if(complex_abs(value+j+1) > complex_abs(value+j))
         {
            temp_real=diff[j];  /* Elemente von diff vertauschen */
            diff[j] = diff[j+1];
            diff[j+1] = temp_real;
            temp_real=value[j].re;      /* Realteil von value vertauschen */
            value[j].re = value[j+1].re;
            value[j+1].re = temp_real;
            temp_real=value[j].im;      /* Imaginaerteil von value vertauschen */
            value[j].im = value[j+1].im;
            value[j+1].im = temp_real;
            temp_int = x[j].el_cnt;     /* Vektoren vertauschen */
            x[j].el_cnt = x[j+1].el_cnt;
            x[j+1].el_cnt = temp_int;
            temp_complex_ptr = x[j].elements;
            x[j].elements = x[j+1].elements;
            x[j+1].elements = temp_complex_ptr;
         }
#endif

#ifdef EXPLICIT
     for(j=cur_pos; (j<nump) && (j < cur_pos+2); j++)
#else
     j = cur_pos;
#endif
      {
       if(! (pass % 10))
        {
         add(value+j, stabilize, &temp_complex);
         printf("%d. Estimate for Eigenvalue %d: (%17.17g) ", pass, j+1, diff[j]);
         show_complex(&temp_complex);
        }
      }

     pass++;
    }
     tv = z;				/* x und z vertauschen */
     z = x;
     x = tv;
     stabilize->im *= -1;
     value[cur_pos].im *= -1;
    }

   for(j=0; j<num; j++)
    {
      add(value+j, stabilize, &temp_complex);
      fprintf(outfile, "e%d := ", j);			/* Ergebnisse in Output-File schreiben */
      output_complex(&temp_complex, outfile);
      fprintf(outfile, ";\n");
      sprintf((char *) hstring, "e%d", j);		/* Eigenwerte speichern */
      assign_variable(hstring, &temp_complex);
      if(j)
       {
         sub(value+j, value, &temp_complex);		/* Eigenwertdifferenzen speichern */
         sprintf((char *) hstring, "d%d", j);
         assign_variable(hstring, &temp_complex);
       }
    }
   for(j=0; j<nump; j++)
    {
     if((store_eigen) && (j < num))
      {
        sprintf((char *) hstring, "ev%d", j);
        alloc_vector(hstring, x[j].el_cnt);
        temp_vec = identify_vector(hstring, TRUE);
        copy_vector(x+j, temp_vec);
        normalize_vector(temp_vec);			/* Sicherheitshalber den Vektor nochmal auf 1 normieren */
      }
     free_vector(x+j);
     free_vector(z+j);
     free_vector(y+j);
    }
free_it:
   if(x)
     free(x);
   if(z)
     free(z);
   if(y);
     free(y);
   if(value);
     free(value);
   if(diff)
     free(diff);
   return(num);
error:
   display_error("Not enough space for eigenvalues (vectors)");
   goto free_it;
 }

/* Eigenwerte ausrechnen */

/* initialize Lanczos' start vector */

static void init_l_startv(struct vector *y)
 {
   register int j;

   for(j=0; j < y->el_cnt; j++)
    {
         y->elements[j].re = 1.0;
         y->elements[j].im = 0;
    }
   normalize_vector(y);
 
/* To show start vector, remove these comments
show_vector(y);
*/
 }

#define LNAME (UBYTE *) "lanc_matrix"

int calculate_eigenvalues(REAL precision, int num, int anum, COMPLEX *stabilize, int lanc_flag,
                          int lanc_size, int big_flag, int non_herm, int jacobi_flag, FILE *outfile)
 {
   int lnum, eign, dim;
   COMPLEX alpha, beta;
   COMPLEX temp_complex;
   struct vector *y, v, *eptr;
   struct vector *temp_vec;
   struct matrix *mat, *matrix;
#ifdef LAPACK
   COMPLEX tempc, tempx;
   REAL *dvec, *evec, *zmat, *work;
   char jobz='I';
   integer info;
#else
   struct vector eig;
   struct matrix *mptr;
   struct full_matrix bchg;
#endif
   register int i, j;
   UBYTE hstring[200];

   if(! matrix_list)
    {
       display_error("No matrix defined");
       return(-1);
    }
   matrix = matrix_list;
   if(matrix_list->next)
    {
       for(matrix = matrix_list; matrix->next; matrix=matrix->next)
                   ;
       printf("*** Warning: More than one matrix defined ***\n");
       printf("*** Using %s ***\n", matrix->text);
    }

   if(non_herm)
    {                                   /* Direkt Eigenwerte berechnen */
     num = non_herm_matrix_eigenvalues(matrix, precision, num, stabilize, outfile);
     return(num);
    }


   if(! lanc_flag)
    {					/* Direkt Eigenwerte berechnen */
     num = matrix_eigenvalues(matrix, precision, num, anum, stabilize, jacobi_flag, outfile);
     return(num);
    }

   printf("Performing Lanczos' prediagonalization\n");

   if(lanc_size > 0)
     lnum = lanc_size;
   else
     lnum = 10*num;			/* Zahl der Lanczos' - Vektoren festlegen */
   if(lnum > matrix->columns)	/* Mehr Eigenvektoren als Spalten gibt es nicht */
      lnum = matrix->columns;

   printf("(using a Krylov subspace of dimension %d)\n", lnum);

   y=0;					/* Vektoren noch nicht allokatiert */
   y=(struct vector *)malloc(lnum * sizeof(struct vector));
   if(! y)
     goto error;
   mat = matrix;
   dim = mat->columns;
   alloc_int_vector(&v, dim);

   if(! big_flag)			/* Alle, oder nur einen Vektor allokatieren ? */
    {
     for(j=0; j<lnum; j++)
       alloc_int_vector(y+j, dim);
    }
   else
    {
      alloc_int_vector(y, dim);
      if(lnum > 1)
        alloc_int_vector(y+1, dim);
    }

   init_l_startv(y);

#ifdef LAPACK
   dvec = (REAL *)malloc(lnum * sizeof(REAL));
   if(! dvec)
     goto lerror;
   evec = (REAL *)malloc((lnum-1) * sizeof(REAL));
   if(! evec)
    {
     free(dvec);
     goto lerror;
    }
#else
   alloc_matrix(LNAME, lnum, lnum);
#endif

   for(j=0; j<lnum; j++)		/* Nun die eigentliche Iteration */
    {
      apply_matrix(mat, y+j, &v);
      if(j)
        sub_vector(&beta, y+j-1, &v);
      sprod(y+j, &v, &alpha);
#ifdef LAPACK
      dvec[j] = alpha.re;
      if(! is_zero(alpha.im/(alpha.re?alpha.re:1.0L)))
        display_error("Lanczos-matrix: diagonal element is not real");
#else
      assign_matrix_el(LNAME, j+1, j+1, &alpha);
#endif
      if(j < lnum-1)
       {
	 if(big_flag)			/* Should we reuse an old vector */
           if(j >= 1)
               move_int_vector(y+j-1, y+j+1);
         copy_vector(&v, y+j+1);
         sub_vector(&alpha, y+j, y+j+1);
         beta.re = norm(y+j+1);
         beta.im = 0.0;
         if(! big_flag)
          {
#ifdef DEBUGDIAG
            debug_out("Gram-Schmidt called");
#endif
            gr_sch1(y, j+2);		/* Zur Sicherheit orthonormieren */
#ifdef DEBUGDIAG
            debug_out("Gram-Schmidt finished");
#endif
          }
         else
            normalize_vector(y+j+1);	/* oder wenigstens auf 1 normieren */
#ifdef LAPACK
         evec[j] = beta.re;
#else
         assign_matrix_el(LNAME, j+1, j+2, &beta);
         assign_matrix_el(LNAME, j+2, j+1, &beta);
#endif
       }
      printf("%d %%\n", (100*j)/lnum);
    }

   printf(" ... done\n");

#ifdef LAPACK
   printf("Diagonalizing now the resulting tridiagonal matrix using dsteqr()\n");
   if(lanc_ev_flag)
    {
     work = (REAL *)malloc(2*lnum * sizeof(REAL));
     if(! work)
      {
       free(dvec);
       free(evec);
       goto lerror;
      }
     zmat = (REAL *)malloc(lnum*lnum * sizeof(REAL));
     if(! zmat)
      {
       free(work);
       free(dvec);
       free(evec);
       goto lerror;
      }
     jobz = 'I';
     dsteqr_(&jobz, &lnum, dvec, evec, zmat, &lnum, work, &info);
     free(work);
    }
   else
    {
     work = zmat = 0;
     jobz = 'N';
     dsteqr_(&jobz, &lnum, dvec, evec, zmat, &lnum, work, &info);
    }
   if(info < 0)
    {
      sprintf((char *) hstring, "Illegal argument %d for dsteqr", -info);
      display_error(hstring);
    }
   if(info > 0)
      display_error("Tridiagonalization did not converge");
   printf(" ... done\n");
   tempx.re = dvec[0];
   tempx.im = 0;
   for(j=0; j<((lnum<num)?lnum:num); j++)
    {
     fprintf(outfile, "e%d := %17.17lg;\n", j, dvec[j]);
     sprintf((char *) hstring, "e%d", j);	   /* store eigenvalues */
     tempc.re = dvec[j];
     tempc.im = 0;
     assign_variable(hstring, &tempc);
     if(j)
      { 					   /* ... and differences */
       sub(&tempc, &tempx, &temp_complex);
       sprintf((char *) hstring, "d%d", j);
       assign_variable(hstring, &temp_complex);
      }
    }
#else
   for(mptr=matrix_list; mptr; mptr=mptr->next)
     if(compare(mptr->text, LNAME))
      {
/* To show resulting tri-diagonal matrix activate this line
show_matrix(mptr);
*/
        printf("Diagonalizing now the resulting tridiagonal matrix\n");
        if(lanc_ev_flag)
         {
          alloc_full_matrix(&bchg, lnum, lnum, FALSE);	/* Init basis change to identity */
          for(i=0; i<lnum; i++)
             bchg.entries[i][i].re = ((REAL) 1);
          tridiag_ql_implicit(mptr, &bchg, &eig);
         }
        else
          tridiag_ql_implicit(mptr, 0, &eig);
        printf(" ... done\n");
        for(j=0; j<((lnum<num)?lnum:num); j++)
         {
          fprintf(outfile, "e%d := %17.17lg;\n", j, (&eig)->elements[j].re);
          sprintf((char *) hstring, "e%d", j);		/* store eigenvalues */
          assign_variable(hstring, (&eig)->elements+j);
          if(j)
           {						/* ... and differences */
            sub((&eig)->elements+j, (&eig)->elements, &temp_complex);
            sprintf((char *) hstring, "d%d", j);
            assign_variable(hstring, &temp_complex);
           }
         }
        free_vector(&eig);
        goto lanc_ev;
      }
    display_error("Lanczos' matrix not found");
    goto free_vec;
#endif

 lanc_ev:
   if(lanc_ev_flag)
    {
      eign = (lnum<num)?lnum:num;
      eptr = (struct vector *)malloc(eign * sizeof(struct vector));
      if(! eptr)
       {
        display_error("Not enough memory for pointer to eigenvectors of Lanczos' method");
        return(-1);
       }

     for(i=0; i<eign; i++)		/* Allocate eigenvectors */
        alloc_int_vector(eptr+i, dim);

     if(big_flag)
      {
       printf("Performing second Lanczos' pass to compute eigenvector(s)\n");
       if(lnum > 2)
         move_int_vector(y+lnum-1, y);
       if(lnum > 3)
         move_int_vector(y+lnum-2, y+1);
       init_l_startv(y);
      }
     /* Accumulate lowest eigenvector in ev */
     for(j=0; j<lnum; j++)
      {
#pragma omp parallel for private(i, tempc)
        for(i=0; i<eign; i++)
	 {
#ifdef LAPACK
          tempc.re = zmat[i*lnum+j];
	  tempc.im = 0;
	  add_vector(&tempc, y+j, eptr+i);
#else
          add_vector(bchg.entries[j]+i, y+j, eptr+i);
#endif
	 }
        if(big_flag)			/* Second Lanczos' pass */
         {
          apply_matrix(mat, y+j, &v);
          if(j)
            sub_vector(&beta, y+j-1, &v);
          sprod(y+j, &v, &alpha);
          if(j < lnum-1)
           {
            if(j >= 1)
                 move_int_vector(y+j-1, y+j+1);
            copy_vector(&v, y+j+1);
            sub_vector(&alpha, y+j, y+j+1);
            beta.re = norm(y+j+1);
            beta.im = 0.0;
            normalize_vector(y+j+1);	/* normalize to 1 */
           }
          printf("%d %%\n", (100*j)/lnum);
         }
      }
     if(mat->text[0] == '*')		/* Internal matrix application ? */
        compute_correl(eptr, outfile);
     dump_vectors(eign, eptr[0].el_cnt, eptr);			/* Write vectors to file (?) */
     if(store_eigen)			/* Store eigenvectors */
       for(i=0; i<eign; i++)
        {
         sprintf((char *) hstring, "ev%d", i);
         alloc_vector(hstring, eptr[i].el_cnt);
         temp_vec = identify_vector(hstring, TRUE);
         copy_vector(eptr+i, temp_vec);
         normalize_vector(temp_vec);	/* Normalize the result to one */
        }
     for(i=0; i<eign; i++)
       free_vector(eptr+i);
     free(eptr);
#ifdef LAPACK
     free(zmat);
#else
     free_full_matrix(&bchg);	/* Free basis change matrix */
#endif
    }
 free_vec:
#ifdef LAPACK
   free(evec);
   free(dvec);
#endif
   free_vector(&v);		/* Vektoren freigeben */

   if(! big_flag)		/* Alle, oder nur einen Vektor freigeben ? */
    {
     for(j=0; j<lnum; j++)
       free_vector(y+j);
    }
   else
    {
       if(lnum > 1)
         free_vector(y+lnum-2);
       free_vector(y+lnum-1);
    }

 free_it:
   if(y)			/* Vektor-Liste freigeben */
     free(y);
   return((lnum<num)?lnum:num);
 error:
   display_error("Not enough memory for Lanczos' vectors");
   goto free_it;
 lerror:
   display_error("Not enough memory for LaPack data in Lanczos' routine");
   goto free_it;
 }

/* *************************************
   Routines for fast reading of matrix */

static int use_buffer = TRUE;			/* Do we want to use a buffer ? */
 
#define NBUFEXPR 1024
 
static UBYTE *buf_expr[NBUFEXPR];
static COMPLEX buf_result[NBUFEXPR];
 
static int buf_cnt = 0;
 
/* Name of matrix under allocation */

static UBYTE cur_mname[256] = "";
static int cur_mname_length = 0;

/* Reset buffer */
 
static void reset_ev_buf()
 {
  register int i;

#ifdef DEBUGBUF
  if(cur_mname_length)
   {
    printf("Buffered the following %d expressions while reading matrix '%s':\n",
           buf_cnt, cur_mname);
    for(i=0; i<buf_cnt; i++)
     {
      printf("\t%s = ", buf_expr[i]);
      show_complex(&(buf_result[i]));
     }
   }
#endif
 
  for(i=0; i<buf_cnt; i++)
     free(buf_expr[i]);
  buf_cnt = 0;

  cur_mname_length = 0;
  cur_mname[0] = 0;
 }

/* Reset notification */

static void reset_matrix_notify()
 {
  reset_ev_buf();
  notify_ptr = void_notify_matrix;
  notify_ptr_overwrite = void_notify_matrix_overwrite;
 }

/* Catch notification for matrix allocation */

static void catch_notify_matrix(UBYTE *text)
 {
  register int i;

  for(i=0; (i<256) && (cur_mname[i] = text[i]); i++)
              ;
  cur_mname_length = i;
 }

/* Catch overwriting of matrix element */

static void catch_matrix_overwrite()
 {
  reset_ev_buf();
 }

/* Buffered evaluation of expression */

static void buf_evaluate(UBYTE *string, COMPLEX *result)
 {
  UBYTE str[STRSIZE];
  register int i, j;

  j=0;				/* Copy, eliminating spaces */
  for(i=0; string[i]; i++)
   if(string[i] > ' ')
     str[j++] = conv_to_lower(string[i]);
  str[j]=0;

  for(i=0; i<buf_cnt; i++)
   if(! strcmp((char *) buf_expr[i], (char *) str))
    {
      result->re = buf_result[i].re;
      result->im = buf_result[i].im;
      return;
    }

  evaluate(str, result);

  if(buf_cnt < NBUFEXPR)
   {
    if(buf_expr[buf_cnt] = (UBYTE *) malloc(strlen((char *) str)+1))
     {
      for(i=0; buf_expr[buf_cnt][i] = str[i]; i++)
              ;
      buf_result[buf_cnt].re = result->re;
      buf_result[buf_cnt].im = result->im;
      buf_cnt++;
     }
    else
     display_error("Failed to allocate buffer memory");
   }
  else
   display_error("Out of buffer space");
 }

/* Files abarbeiten */

int process_files(char *infile, char *outfile, REAL precision, int num, int anum, COMPLEX *stabilize, int lanc_flag,
                  int lanc_size, int big_flag, int non_herm, int jacobi_flag)
 {
   FILE *fpi, *fpo;
   int ofs;
   register int i, j;
   UBYTE instr[STRSIZE], lbuf[STRSIZE];
   COMPLEX result;
   int row, column;
   int iflag;

   if(infile[0] == '*')
    {
     alloc_matrix((UBYTE *) infile, 0, 0);
     fpo = zip_open(outfile, "w");
     if(! fpo)
      {
        display_error_text("Failed to open '%s' for output", outfile);
        return(-1);
      }
     num = calculate_eigenvalues(precision, num, anum, stabilize, lanc_flag, lanc_size, big_flag, non_herm, jacobi_flag, fpo);
     fprintf(fpo, "\n\n;end;\n");
     zip_close(fpo, "w");
     return(num);
    }

   fpi=zip_open(infile, "r");
   if(! fpi)
    {
      display_error_text("Failed to open '%s' for input", infile);
      return(-1);
    }
   fpo=zip_open(outfile, "w");
   if(! fpo)
    {
      zip_close(fpi, "r");
      display_error_text("Failed to open '%s' for output", outfile);
      return(-1);
    }

   /* Catch critical events */
   if(use_buffer)		/* Use buffer ? */
    {
     notify_ptr = catch_notify_matrix;
     notify_ptr_overwrite = catch_matrix_overwrite;
    }

   ofs = 0;
   printf("Reading '%s'\n", infile);
   while(1)
    {
     if(! fgets((char *) lbuf, STRSIZE-1, fpi))
      {
        display_error("File reading not properly terminated");
        if(use_buffer)	/* Use buffer ? */
          reset_matrix_notify();
        return(num); 
      }
     for(i=0; (i<2048) && (lbuf[i] <= ' ') && lbuf[i]; i++)
                     ;
     for(; (i<2048) && (lbuf[i] != '%') && lbuf[i]; i++)
       if(lbuf[i] >= ' ')
        {
         if((lbuf[i] == ';') || (lbuf[i] == '$'))
          {
           if(! ofs)
            {
             if(! strncasecmp((char *) lbuf+i+1, "end;", 4)) 
              {
               printf(" ... done\n");
               num = calculate_eigenvalues(precision, num, anum, stabilize, lanc_flag, lanc_size, big_flag, non_herm, jacobi_flag, fpo);
               zip_close(fpi, "r");
               fprintf(fpo, "\n\n;end;\n");
               zip_close(fpo, "w");
               if(use_buffer)	/* Use buffer ? */
                 reset_matrix_notify();
               return(num);
              }
            }
           else
            {
             instr[ofs] = 0;
             ofs = 0;
             iflag = TRUE;
             if(cur_mname_length)
              if(instr[cur_mname_length] == '(')
               if(! strncasecmp((char *) instr, (char *) cur_mname, cur_mname_length))
                {		/* Matches current matrix */
                 for(j=cur_mname_length+1; instr[j] && (instr[j] <= ' '); j++)
                        ;	/* Next non-space character */
                 j = get_number(instr, j, &row);
                 for(; instr[j] && (instr[j] <= ' '); j++)
                        ;	/* Next non-space character */
                 if(instr[j++] == ',')
                  {		/* must be a comma for a matrix */
                   j = get_number(instr, j, &column);
                   for(; instr[j] && (instr[j] <= ' '); j++)
                          ;	/* Next non-space character */
                   if(instr[j++] == ')')
                    {		/* must be a closing bracket */
                     for(; instr[j] && (instr[j] <= ' '); j++)
                          ;	/* Next non-space character */
                       if(instr[j++] == ':')
                         if(instr[j++] == '=')
                          {	/* ":=" must follow */
                            for(; instr[j] && (instr[j] <= ' '); j++)
                                ; 	/* Next non-space character */
                            if(use_buffer)	/* Use buffer ? */
                              buf_evaluate(instr+j, &result);
                            else
                              evaluate(instr+j, &result);
/*
printf("%s(%d,%d) := %s = ", cur_mname, row, column, instr+j);
show_complex(&result);
*/
                            if(cur_mname_length)
                             {
                              assign_matrix_el(cur_mname, row, column, &result);
                              iflag = FALSE;
                             }
                          }
                    } 
                  }
                }
             if(iflag)
              {
/*
printf("Interpreting: %s\n", instr);
*/
               if(cur_mname_length && use_buffer)
                      reset_ev_buf();
               interpret_line(instr, &result, FALSE);
              }
            }
          }
         else
           if(ofs || lbuf[i] > ' ')
             instr[ofs++] = lbuf[i];
        }
    }
   if(use_buffer)	/* Use buffer ? */
     reset_matrix_notify();
   return(num);
 }

#ifdef READLINE

/* A static variable for holding the line. */
static char *line_read = (char *)NULL;

/* Read a string using GNU readline, and return a pointer to it.
   Returns NULL on EOF. */
char *rl_gets(char *prompt)
 {
  /* If the buffer has already been allocated,
     return the memory to the free pool. */
  if(line_read)
    {
      free(line_read);
      line_read = (char *)NULL;
    }

  /* Get a line from the user. */
  line_read = readline(prompt);

  /* If the line has any text in it,
     save it on the history. */
  if(line_read && *line_read)
    add_history(line_read);

  return(line_read);
 }
#endif

/* Show options */

#ifdef _OPENMP
int omp_numproc;
#endif

void show_options()
 {
   printf("%s\nDate: %s\n", COPYRIGHT, DATE);
#ifdef _OPENMP
   omp_numproc = omp_get_max_threads();
   printf("This version is parallelized using OpenMP ... using %d processors\n",
          omp_numproc);
#endif
#ifdef LAPACK
   printf("... uses LaPack instead of Householder algorithm for matrix diagonalization\n");
#endif
 }

/* Das Hauptprogramm */

main(int argc, char *argv[])
 {
   COMPLEX result;
   int ofs, cnt, pc_flag, cont_flag, lanc_flag, lanc_size, big_flag, non_herm, jacobi_flag;
   register int i, j;
   UBYTE string[256], *istring;
   UBYTE hstring[100];
   char *infile, *outfile;
   REAL precision;
   COMPLEX stabilize;
   int num, anum;
   REAL tnum;

   cnt = 0;
   for(i=0; *(argv[0] + i); i++)
     if((*(argv[0]+i) == '/') || (*(argv[0]+i) == '\\'))
         cnt = i+1;
   *(argv[0] + cnt) = 0;
   set_helppath(argv[0]);

   precision = 1e-12;
   anum = 2;
   num = 10;
   stabilize.re = stabilize.im = 0.0;

#ifdef SSE
   init_sse();
#endif
   init_short_malloc();

   /* Copy dump file name */
   for(i=0; VIDMPNAMEA[i] = VIDMPNAMEB[i] = VIDMPNAME[i]; i++)
            ;

   use_buffer = TRUE;			/* By default we want to buffer input files while reading them */
   sprod_complex = TRUE;
   read_before = write_after = FALSE;
   skip_first = FALSE;			/* Default: Always do a diagonalization pass first */
   no_eigenvects = FALSE;		/* Default: We want eigenvectors */
   mel_flag = FALSE;			/* Default: do not compute matrix elements */
   gzip_out = FALSE;			/* Default: uncompressed output */
   gzip_flag = jacobi_flag = non_herm = store_eigen = big_flag=lanc_flag=cont_flag=pc_flag=FALSE;
   lanc_size = -1;
   infile=outfile=0;
   if(argc <= 1)
      goto display_help;
   for(i=1; i<argc; i++)
    {
      if(*argv[i] == '-')
       {
         switch(*(argv[i]+1))
          {
            case 'r':
            case 'R':
              sprod_complex = FALSE;
              break;
            case 'a':
            case 'A':
	      if((argv[i][2] | 0x20) == 'l')
	        no_eigenvects = TRUE;
	      else
	       {
                sscanf(argv[i]+2, "%lg", &tnum);
                precision = tnum;
	       }
              break;
            case 'n':
            case 'N':
              sscanf(argv[i]+2, "%d", &num);
              for(j=2; (argv[i][j]) && (argv[i][j] != '+'); j++)
                         ;
              if(argv[i][j])
                sscanf(argv[i]+j+1, "%d", &anum);
              break;
            case 'p':
            case 'P':
              pc_flag=TRUE;
              break;
            case 'c':
            case 'C':
              cont_flag=TRUE;
              break;
            case 'l':
            case 'L':
              lanc_flag=TRUE;
              ofs=2;
              if((*(argv[i]+ofs) | 0x20) == 'e')
               {
                ofs++;
                lanc_ev_flag = TRUE;
               }
              if(*(argv[i]+ofs) == '=')
                sscanf(argv[i]+ofs+1, "%d", &lanc_size);
              break;
            case 'b':
            case 'B':
              big_flag=TRUE;
              break;
            case 'h':
            case 'H':
             non_herm = TRUE;
             break;
            case 'j':
            case 'J':
             jacobi_flag = TRUE;
             break;
            case 'z':
            case 'Z':
             if((argv[i][2] | 0x20) == 'o')
               gzip_out = TRUE;
             else
               gzip_flag = TRUE;
             break;
            case 'd':
            case 'D':
             use_buffer = FALSE;
             break;
            case 'F':
             skip_first = TRUE;
            case 'f':
             read_before = TRUE;
             if(*(argv[i]+2))				/* Copy name */
               for(j=0; VIDMPNAMEB[j] = *(argv[i]+j+2); j++)
                       ;
             break;
            case 't':
            case 'T':
             write_after = TRUE;
             if(*(argv[i]+2))				/* Copy name */
               for(j=0; VIDMPNAMEA[j] = *(argv[i]+j+2); j++)
                       ;
             break;
            case 'e':
            case 'E':
              store_eigen=TRUE;
              break;
            case 'm':
            case 'M':
	      switch(argv[i][2] | 0x20)
	       {
	        case 'a':
                 mel_flag=ALLMEL;
		 break;
	        case 'd':
                 mel_flag=DIAGMEL;
		 break;
	        case 'u':
                 mel_flag=ABOVEMEL;
		 break;
                default:
		 display_error("Use '-m' either as '-ma', '-mu' or '-md'");
                 goto display_help;
	       }
              break;
            case 'i':
            case 'I':
              interpret_line((UBYTE *) (argv[i]+2), &result, FALSE);
              break;
            case 's':
            case 'S':
              interpret_line((UBYTE *) (argv[i]+2), &stabilize, FALSE);
              break;
            default:
              display_error_text("Invalid option '%s'", argv[i]);
              goto display_help;
          }
       }
      else
       {
        if(infile && outfile)
         {
          display_error("Too many files specified");
          goto display_help;
         }
        if(! infile)
          infile = argv[i];
        else
          outfile = argv[i];
       }
    }

   if(! pc_flag)
    {
      printf("Evaluation of matrix eigenvalues\n");
      show_options();
      printf("\n");
      if(! outfile)
       {
         display_error("Not enough files specified");
         goto display_help;
       }
      num = process_files(infile, outfile, precision, num, anum, &stabilize, lanc_flag, lanc_size, big_flag, non_herm, jacobi_flag);
      if(! cont_flag)
       {
        output_close();
        free_variables();
        free_matrices();
        free_vectors();
        free_functions();
        exit(0);
       }
      printf("\nEigenvalues are stored in e0 .. e%d\n", num-1);
      printf("e(i)-e(0) are stored in d1 ... d%d\n", num-1);
      if(store_eigen)
       {
        if(lanc_flag)
          printf("The extremal eigenvector of the Lanczos method is stored in ev0\n");
        else
          printf("The corresponding eigenvectors of the power method are stored in ev0 .. ev%d\n", num-1);
       }
      printf("\n");
    }

   printf("Pocket-calculator\n");
   show_options();
   printf("\nFor further information type 'help'\n\n");
   printf("Enter expression to be evaluated\n");
   printf("or 'bye' (or 'quit') to quit\n\n");
   cnt=1;
   istring = string;
loop:
#ifdef READLINE   
   sprintf((char *) string, "%d: ", cnt);
   istring = (UBYTE *) rl_gets((char *) string);
#else
   printf("%d: ", cnt);
   /* Use fgets() instead of gets() to avoid buffer overflow */
   fgets((char *) string, 255, stdin);
   /* remove "dirt" at the end of the line */
   for(i=0; (i<256) && string[i]; i++)
   		;		/* find end */
   for(; (i>=0) && (string[i] <= ' '); i--)
          string[i] = 0;
#endif
   if(! istring[0])		/* Keine leeren Befehle speichern */
     goto loop;
   interpret_line(istring, &result, TRUE);
   sprintf((char *) hstring, "%%%d", cnt);	/* Ergebnis speichern */
   assign_variable(hstring, &result);
   cnt++;
   goto loop;

display_help:			/* Hilfe ausgeben */
   printf("You must specify at least one of the following options:\n");
   printf("-p         Start as pocket calculator\n");
   printf("-c         Continue with pocket calculator after evaluation of eigenvalues\n");
   printf("-z         Read input from gzip'ped file ('.gz' is automatically appended\n");
   printf("           to filename). Option only effective if not in pocket-calculator mode.\n");
   printf("-zo        Compress output file using gzip ('.gz' is automatically appended\n");
   printf("           to filename).\n");
   printf("-d         Direct evaluation of matrix elements while reading input (use if\n");
   printf("           more than %d different entries in an input matrix)\n", NBUFEXPR);
   printf("-l         Diagonalize using Lanczos' method\n"); /* before applying power method\n"); */
   printf("-l=num     Same as -l, but explicitly specify dimension of subspace to equal num\n");
   printf("-le[=num]  Same as -l, -l=num, but compute also eigenvectors with Lanczos'\n");
   printf("           method\n");
   printf("-b         Suppress Gram-Schmidt orthogonalization during Lanczos' iteration\n");
   printf("           (Use for very big matrices)\n");
   printf("-h         Vector iteration for non-hermitean matrices\n");
   printf("-j         Use Jacobi method for explicit diagonalization step\n");
   printf("           (otherwise Householder is used by default)\n");
   printf("-e         Store eigenvectors of power method\n");
   printf("-m         '-ma' computes ALL, '-md' the DIAGONAL and '-mu' the upper elements\n");
   printf("           of matrices other than the Hamiltonian in the eigenbasis and stores\n");
   printf("           them in the output file\n");
   printf("-r         Suppress complex conjugation in scalar product\n");
   printf("           (Intended for eigenvalues of non-hermitian but symmetric matrices)\n");
   printf("-f[file]   Read vectors for vector iteration from dump 'file' before starting\n");
   printf("           iteration (uses '%s' if 'file' is omitted)\n", VIDMPNAME);
   printf("-F[file]   same as -f, except that it tries to skip the first diagonalization\n");
   printf("            pass\n");
   printf("-t[file]   Save vectors for vector iteration to dump 'file' at the end of the\n");
   printf("           iteration (uses '%s' if 'file' is omitted)\n", VIDMPNAME);
   printf("-anum      Approximate eigenvalues to precision of num (default: 1e-12)\n");
   printf("-all       Compute ALL eigenvalues to machine precision. Saves CPU time by NOT\n");
   printf("           computing eigenvectors (i.e. -e is ignored and accuracy not checked)\n");
   printf("-nN[+M]    Approximate N eigenvalues (default: 10)\n");
   printf("           iterate M more than needed (default: 2)\n");
   printf("-sexprn    Stabilize iteration by shifting eigenvalues by -exprn\n");
   printf("-istring   Interpret string via pocket calculator\n");
   printf("If not in pocket-calculator mode you must specify two\n");
   printf("filenames. The first will be taken for input, the second\n");
   printf("for output\n");
   printf("REMARK: During power method, the submatrices are explicitly diagonalized\n");
   printf("        using the Householder (or Jacobi) method in order to improve\n");
   printf("        separation of eigenvalues\n");
 }
