/* Definitions for matrices */

/* A matrix-element */

#ifdef SSE	/* optimized for alignment of complex component */
struct matrix_el { COMPLEX value;
                   struct matrix_el *next;
                   int column;
                 };
#else
struct matrix_el { struct matrix_el *next;
                   int column;
                   COMPLEX value;
                 };
#endif

/* A complete matrix */

struct matrix { struct matrix *next;
                UBYTE *text;
                int rows;
                int columns;
                struct matrix_el **matrix_rows;
              };

/* List of all matrices */

extern struct matrix *matrix_list;

/* Functions for matrices */

extern void free_matrix_elements(struct matrix_el *ptr);
extern void free_matrices();
extern void alloc_matrix(UBYTE *text, int rows, int columns);
extern void insert_matrix_el(struct matrix_el **list, int column, COMPLEX *value);
extern void assign_matrix_el(UBYTE *str, int row, int column, COMPLEX *value);
extern struct matrix *identify_matrix(UBYTE *text);
extern void show_matrix(struct matrix *matrix);
extern void adjoint_matrix(struct matrix *src, struct matrix *dst);
extern int get_matrix_var(UBYTE *instr, UBYTE *str, int *rows, int *columns);
extern int get_matrix_el(struct matrix *matrix, int row, int column, COMPLEX *result);
extern void madd(struct matrix *src1, struct matrix *src2, struct matrix *dst);
extern void mprod(struct matrix *src1, struct matrix *src2, struct matrix *dst);
extern void commute(struct matrix *src1, struct matrix *src2, struct matrix *dst);

/* Pointer to function for notification of matrix definition */
extern void (*notify_ptr)(UBYTE *text);
/* Void function for notification of matrix definition */
extern void void_notify_matrix(UBYTE *text);
/* Pointer to function for notification of matrix element overwriting */
extern void (*notify_ptr_overwrite)();
/* Void function for notification of matrix element overwriting */
extern void void_notify_matrix_overwrite();
